import shutil

from pyrefinebio import computed_file as prb_computed_file
from pyrefinebio.api_interface import download_file, get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.exceptions import DownloadError, MissingFile
from pyrefinebio.util import create_paginated_list, expand_path


class Compendium(Base):
    """Compendium

    Retrieve a Compendium based on id

        >>> import pyrefinebio
        >>> id = 1
        >>> result = pyrefinebio.Compendium.get(id)

    Retrieve a list of Compendia based on filters

        >>> import pyrefinebio
        >>> results = pyrefinebio.Compendium.search(primary_organism__name="ACTINIDIA_CHINENSIS")
    """

    def __init__(
        self,
        id=None,
        primary_organism_name=None,
        organism_names=None,
        svd_algorithm=None,
        quant_sf_only=None,
        compendium_version=None,
        computed_file=None,
    ):
        super().__init__(identifier=id)

        self.id = id
        self.primary_organism_name = primary_organism_name
        self.organism_names = organism_names
        self.svd_algorithm = svd_algorithm
        self.quant_sf_only = quant_sf_only
        self.compendium_version = compendium_version
        self.computed_file = (
            prb_computed_file.ComputedFile(**(computed_file)) if computed_file else None
        )

        self._downloaded_path = None

    @classmethod
    def get(cls, id):
        """Retrieve a specific Compendium based on id

        Returns:
            Compendium

        Parameters:
            id (int): the id for the Compendium you want to get
        """
        response = get_by_endpoint("compendia/" + str(id)).json()
        return Compendium(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of Compendium results based on filters

        Returns:
            list of Compendium

        Keyword Arguments:
            primary_organism__name (str): filter based on the name of the primary Organism
                                          associated with the compendium

            compendium_version (int): filter based on the Compendium's version

            quant_sf_only (bool): true for RNA-seq Sample Compendium results or False
                                  for quantile normalized

            result__id (int): filter based on the id of the ComputationalResult associated
                              with the compendium

            ordering (str): which field to use when ordering the results

            limit (int): number of results to return per page

            offset (int): the initial index from which to return the results

            latest_version (bool): true will only return the highest
                                   compendium_version for each primary_organism
        """
        response = get_by_endpoint("compendia", params=kwargs)
        return create_paginated_list(cls, response)

    def download(self, path, prompt=True):
        """Download a Compendium result

        Returns:
            Compendium

        Parameters:
            path (str): the path that the Compendium result should be downloaded to

            prompt (bool): if true, will prompt before downloading files bigger than 1GB
        """
        if not self.computed_file.download_url:
            raise DownloadError(
                "Compendia",
                "Download url not found - make sure you have set up and activated your Token. "
                "You can create and activate a new token using pyrefinebio.create_token(). "
                "See documentation for advanced usage: https://alexslemonade.github.io/refinebio-py/token.html",
            )

        full_path = expand_path(path, "compendium-" + str(self.id) + ".zip")

        download_file(self.computed_file.download_url, full_path, prompt)

        self._downloaded_path = full_path

        return self

    def extract(self):
        """Extract a downloaded Compendium

        Returns:
            Compendium
        """
        if not self._downloaded_path:
            raise MissingFile(
                "Compendium downloaded file",
                "Make sure you have successfully downloaded the Compendium before extracting.",
            )

        shutil.unpack_archive(self._downloaded_path)
        return self



from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.util import create_paginated_list


class Processor(Base):
    """Processor.

    Retrieve a Processor by id

        >>> import pyrefinebio
        >>> id = 1
        >>> processor = pyrefinebio.Processor.get(id)

    Retrieve a list of Processors

        >>> import pyrefinebio
        >>> processors = pyrefinebio.Processor.search()
    """

    def __init__(self, id=None, name=None, version=None, docker_image=None, environment=None):
        super().__init__(identifier=id)

        self.id = id
        self.name = name
        self.version = version
        self.docker_image = docker_image
        self.environment = environment

    @classmethod
    def get(cls, id):
        """Retrieve a Processor based on id

        Returns:
            Processor

        Parameters:
            id (int): the id for the Processor you want to get
        """
        response = get_by_endpoint("processors/" + str(id)).json()
        return Processor(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of Processors

        Returns:
            list of Processor

        Since there are no filters, this method always returns all Processors
        """
        response = get_by_endpoint("processors", params=kwargs)
        return create_paginated_list(cls, response)



import json
from datetime import timedelta

import click
import pyrefinebio.high_level_functions as hlf
import pytimeparse
from pyrefinebio.exceptions import DownloadError


@click.group()
def cli():
    """RefineBio CLI"""
    pass


class JsonParamType(click.ParamType):
    name = "json"

    def convert(self, value, param, ctx):
        try:
            return json.loads(value)
        except:
            self.fail("expected valid json string, got: {0}".format(value), param, ctx)


class ListParamType(click.ParamType):
    name = "list"

    def convert(self, value, param, ctx):
        try:
            return value.split()
        except:
            self.fail("expected valid string in list form, got: {0}".format(value), param, ctx)


class TimedeltaParamType(click.ParamType):
    name = "timedelta"

    def convert(self, value, param, ctx):
        try:
            seconds = pytimeparse.parse(value)
            return timedelta(seconds=seconds)
        except:
            self.fail(
                "expected a string representing a timeout like '5 minutes', got: {0}".format(value),
                param,
                ctx,
            )


@cli.command(no_args_is_help=True)
@click.argument("entity", nargs=1)
def describe(entity=None):
    """
    Prints out information about pyrefinebio's classes and functions.

    ENTITY is the pyrefinebio class or function you want information about.

    To get information about a class method, pass in the class name and method name
    separated by either a space or a `.`

    Example:

    $ refinebio describe "Sample.search"
    """
    hlf.help(entity)


@cli.command()
@click.option(
    "--email-address",
    required=True,
    help="The email that will be contacted with info related to the Dataset",
)
@click.option(
    "--path", default="./", type=click.Path(), help="Path that the dataset should be downloaded to"
)
@click.option(
    "--dataset-dict",
    default=None,
    type=JsonParamType(),
    help=(
        "A fully formed Dataset `data` attribute in JSON format. Use this parameter if you want "
        'to specify specific Samples for your Dataset. Ex: \'{"GSE74410": ["GSM1919903"]}\''
    ),
)
@click.option(
    "--experiments",
    default=None,
    type=ListParamType(),
    help="A space separated list of experiment accession codes. Ex: 'SRP051449 GSE44421 GSE44422'",
)
@click.option(
    "--aggregation",
    default="EXPERIMENT",
    type=click.Choice(("EXPERIMENT", "SPECIES", "ALL"), case_sensitive=True),
    help="How the Dataset should be aggregated",
)
@click.option(
    "--transformation",
    default="NONE",
    type=click.Choice(("NONE", "MINMAX", "STANDARD"), case_sensitive=True),
    help="The transformation for the Dataset",
)
@click.option(
    "--skip-quantile-normalization",
    default=False,
    type=click.BOOL,
    help="Control whether the Dataset should skip quantile normalization for RNA-seq Samples",
)
@click.option(
    "--timeout",
    default=None,
    type=TimedeltaParamType(),
    help="A string representing how long to wait for the dataset to process such as '15 minutes'.",
)
@click.option(
    "--notify-me",
    is_flag=True,
    help="Control whether or not refine.bio should send you an email when your Dataset has finished processing.",
)
def download_dataset(
    email_address,
    path,
    dataset_dict,
    experiments,
    aggregation,
    transformation,
    skip_quantile_normalization,
    timeout,
    notify_me,
):
    """
    Automatically constructs a Dataset, processes it, waits for it
    to finish processing, then downloads it to the path specified.
    """
    try:
        hlf.download_dataset(
            path,
            email_address,
            dataset_dict,
            experiments,
            aggregation,
            transformation,
            skip_quantile_normalization,
            timeout=timeout,
            notify_me=notify_me,
        )
    except DownloadError as e:
        raise click.ClickException(str(e))


@cli.command()
@click.option("--organism", help="The name of the Organism for the Compendium you want to download")
@click.option(
    "--path",
    default="./",
    type=click.Path(),
    help="Path that the Compendium should be downloaded to",
)
@click.option(
    "--quant-sf-only",
    default=False,
    type=click.BOOL,
    help="True for RNA-seq Sample Compendium results or False for quantile normalized",
)
def download_compendium(organism, path, quant_sf_only=False):
    """
    Download a Compendium for the specified organism.
    For more information on normalized Compendia check out the following link:
    http://docs.refine.bio/en/latest/main_text.html#normalized-compendia
    """
    try:
        hlf.download_compendium(path, organism, quant_sf_only)
    except DownloadError as e:
        raise click.ClickException(str(e))


@cli.command()
@click.option("--organism", help="The name of the Organism for the Compendium you want to download")
@click.option(
    "--path",
    default="./",
    type=click.Path(),
    help="Path that the Compendium should be downloaded to",
)
def download_quantfile_compendium(organism, path):
    """
    Download a Compendium for the specified organism.
    This command will always download RNA-seq Sample Compendium results.
    For more information on RNA-seq Sample Compendia check out the following link:
    http://docs.refine.bio/en/latest/main_text.html#rna-seq-sample-compendia
    """
    try:
        hlf.download_quantfile_compendium(path, organism)
    except DownloadError as e:
        raise click.ClickException(e.message)


@cli.command()
@click.option(
    "-s",
    "--silent",
    is_flag=True,
    help="Add this flag if you don't want to be prompted before activating AND saving your token.",
)
def create_token(silent):
    """
    Automatically creates a Token, activates it, and stores it to the Config file.
    The Config file's location is `~/.refinebio.yaml` this file will be created if it
    doesn't exist. For more information see pyrefinbio.Config.
    https://alexslemonade.github.io/refinebio-py/config.html

    By default, will prompt the user before activating and storing the created Token.
    """
    agree_to_terms = silent or None
    save_token = silent or None
    hlf.create_token(agree_to_terms, save_token)



class ServerError(Exception):
    base_message = "The server encountered an issue"
    def __init__(self, message=None):
        if message:
            self.base_message = message
        super().__init__(self.base_message)


class BadRequest(Exception):
    base_message = "Bad Request"
    def __init__(self, message=None):
        if message:
            self.base_message += ": " + str(message)
        super().__init__(self.base_message)


class NotFound(Exception):
    base_message = "The requested resource at {0} was not found"
    def __init__(self, url):
        super().__init__(self.base_message.format(url))


class InvalidFilters(Exception):
    base_message = (
        "You have provided invalid filters: {0}\n"
        "Check the documentation for a list of valid filters\n"
        "https://alexslemonade.github.io/refinebio-py/"
    )
    def __init__(self, invalid_filters):
        super().__init__(self.base_message.format(invalid_filters))


class InvalidFilterType(Exception):
    base_message = "You have provided invalid type for the filter, {0} - {1}"
    def __init__(self, filter, info):
        super().__init__(self.base_message.format(filter, info))


class InvalidData(Exception):
    base_message = ""
    def __init__(self, message, details):
        self.base_message = message
        if details:
            self.base_message += "\nDetails: " + str(details)
        super().__init__(self.base_message)


class DownloadError(Exception):
    base_message = "Unable to download {0}"
    def __init__(self, type, extra_info=None):
        if extra_info:
            self.base_message += "\n" + extra_info 
        super().__init__(self.base_message.format(type))


class MultipleErrors(Exception):
    base_message = "Multiple errors have occurred:\n"
    def __init__(self, errors):
        for error in errors:
            self.base_message += str(error) + "\n"
        super().__init__(self.base_message)


class MissingFile(Exception):
    base_message = "Missing file: {0}"
    def __init__(self, file_name, extra_info=None):
        if extra_info:
            self.base_message += "\n" + extra_info
        super().__init__(self.base_message.format(file_name))



import os

import iso8601
from pyrefinebio.api_interface import get


def create_paginated_list(T, response):
    return PaginatedList(T, response)


def parse_date(date):
    try:
        parsed = None
        parsed = iso8601.parse_date(date)
    finally:
        return parsed


def expand_path(path, filename):
    user_path = os.path.expanduser(path)
    full_path = os.path.abspath(user_path)

    if os.path.isdir(full_path):
        full_path = os.path.join(full_path, filename)

    return full_path


class PaginatedList:
    """PaginatedList

    PaginatedLists are returned by all model class `search` methods that deal
    with paginated api responses. PaginatedLists can be indexed like lists and
    iterated through like generators.
    """

    def __init__(self, T, response):
        self.cur = 0
        self.type = T

        self.base_url = response.url

        json = response.json()

        self.total_items = json["count"]
        self.page_size = len(json["results"])

        max_pages = int(self.total_items / self.page_size) if self.page_size else 0

        self.pages = [None for _ in range(max_pages + 1)]
        self.pages[0] = [T(**item) for item in json["results"]]

    def __getitem__(self, index):

        if isinstance(index, slice):
            step = index.step or 1

            default_start = 0 if step > 0 else self.total_items - 1
            default_stop = self.total_items - 1 if step > 0 else 0

            start = index.start or default_start
            stop = index.stop or default_stop

            if start < 0:
                start += self.total_items

            if stop < 0:
                stop += self.total_items

            if start < 0 or stop < 0:
                raise IndexError("index out of range!")

            return [self[i] for i in range(start, stop, step)]

        if index < 0:
            index += self.total_items

        if index >= self.total_items or index < 0:
            raise IndexError("index out of range!")

        page = int(index / self.page_size)
        page_index = index - page * self.page_size

        try:
            result = self.pages[page][page_index]
        except:
            params = {"offset": page * self.page_size, "limit": self.page_size}

            response = get(self.base_url, params=params).json()

            if self.total_items != response["count"]:
                raise RuntimeError("List has changed since creation!")

            self.pages[page] = [self.type(**item) for item in response["results"]]

            result = self.pages[page][page_index]

        return result

    def __setitem__(self, index, value):
        raise AttributeError("PaginatedLists are immutable")

    def __delitem__(self, index, value):
        raise AttributeError("PaginatedLists are immutable")

    def __len__(self):
        return self.total_items

    def __next__(self):
        if self.cur < self.total_items:
            n = self[self.cur]
            self.cur += 1
            return n
        else:
            raise StopIteration()



import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

experiment = {
    "id": 0,
    "title": "test",
    "description": "test experiment",
    "annotations": [
        {
            "data": "test",
            "is_ccdl": True,
            "created_at": "2020-10-02T18:48:24Z",
            "last_modified": "2020-10-02T18:48:24Z",
        }
    ],
    "samples": [
        {
            "accession_code": "test",
            "platform_name": "test",
            "pretty_platform": "test",
            "technology": "test",
            "is_processed": True,
        }
    ],
    "protocol_description": "string",
    "accession_code": "string",
    "alternate_accession_code": "string",
    "source_database": "string",
    "source_url": "string",
    "has_publication": True,
    "publication_title": "string",
    "publication_doi": "string",
    "publication_authors": ["string"],
    "pubmed_id": "string",
    "source_first_published": "2020-10-02T18:48:24Z",
    "source_last_modified": "2020-10-02T18:48:24Z",
    "submitter_institution": "string",
    "last_modified": "2020-10-02T18:48:24Z",
    "created_at": "2020-10-02T18:48:24Z",
    "organism_names": ["string"],
    "sample_metadata": "string",
    "num_total_samples": 1,
    "num_processed_samples": 1,
    "num_downloadable_samples": 1,
}

experiment_search_result_1 = {
    "id": 42,
    "title": "test title 1",
    "publication_title": "test title 1",
    "description": "test description 1",
    "technology": "RNA-SEQ",
    "accession_code": "SRP150473",
    "alternate_accession_code": "GSE115746",
    "submitter_institution": "UPenn",
    "has_publication": True,
    "publication_doi": "",
    "publication_authors": ["Tate S", "Shapiro J"],
    "sample_metadata_fields": ["specimen_part", "disease", "subject"],
    "platform_names": ["Illumina HiSeq 2500"],
    "platform_accession_codes": ["IlluminaHiSeq2500"],
    "organism_names": ["HUMAN"],
    "pubmed_id": "30382198",
    "num_total_samples": 28706,
    "num_processed_samples": 0,
    "num_downloadable_samples": 0,
    "source_first_published": "2018-05-31T00:00:00Z",
}

experiment_search_result_2 = {
    "id": 43,
    "title": "test title 2",
    "publication_title": "test title 2",
    "description": "test description 2",
    "technology": "RNA-SEQ",
    "accession_code": "SRP150473",
    "alternate_accession_code": "GSE115746",
    "submitter_institution": "Drexel",
    "has_publication": True,
    "publication_doi": "",
    "publication_authors": ["Wheeler K", "Mejia D"],
    "sample_metadata_fields": ["specimen_part", "disease", "subject"],
    "platform_names": ["Illumina HiSeq 2500"],
    "platform_accession_codes": ["IlluminaHiSeq2500"],
    "organism_names": ["HUMAN", "GORILLA"],
    "pubmed_id": "30382198",
    "num_total_samples": 28706,
    "num_processed_samples": 0,
    "num_downloadable_samples": 0,
    "source_first_published": "2018-05-31T00:00:00Z",
}

search_1 = {
    "count": 2,
    "next": "search_2",
    "previous": None,
    "results": [experiment_search_result_1],
}

search_2 = {
    "count": 2,
    "next": None,
    "previous": "search_1",
    "results": [experiment_search_result_2],
}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/experiments/SRP150473/":
        return MockResponse(experiment, url)

    if url == "https://api.refine.bio/v1/experiments/bad-accession-code/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/experiments/force-500-error/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/search/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)


class ExperimentTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_experiments_get(self, mock_request):
        result = pyrefinebio.Experiment.get("SRP150473")
        self.assertObject(result, experiment)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_experiments_500(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.ServerError):
            pyrefinebio.Experiment.get("force-500-error")

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_experiments_get_404(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.NotFound):
            pyrefinebio.Experiment.get("bad-accession-code")

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_experiments_search_no_filters(self, mock_request):
        results = pyrefinebio.Experiment.search()

        self.assertObject(results[0], experiment_search_result_1)
        self.assertObject(results[1], experiment_search_result_2)

    def test_experiments_search_with_filters(self):
        filtered_results = pyrefinebio.Experiment.search(
            num_downloadable_samples=20, has_publication="true"
        )  # TODO: investigate why True breaks this endpoint

        for result in filtered_results:
            self.assertEqual(result.num_downloadable_samples, 20)
            self.assertTrue(result.has_publication)



import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

job_1 = {
    "id": 1,
    "source_type": "GEO",
    "success": True,
    "start_time": "2020-10-13T14:38:02.287672Z",
    "end_time": "2020-10-13T14:50:52.962129Z",
    "created_at": "2020-10-13T14:38:02.275537Z",
    "last_modified": "2020-10-13T14:50:52.962141Z",
}

job_2 = {
    "id": 2,
    "source_type": "SRA",
    "success": False,
    "start_time": None,
    "end_time": None,
    "created_at": "2020-10-01T20:12:30.926574Z",
    "last_modified": "2020-10-01T20:12:32.836891Z",
}

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [job_1]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [job_2]}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/jobs/survey/1/":
        return MockResponse(job_1, url)

    if url == "https://api.refine.bio/v1/jobs/survey/2/":
        return MockResponse(job_2, url)

    if url == "https://api.refine.bio/v1/jobs/survey/0/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/jobs/survey/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/jobs/survey/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)


class SurveyJobTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_survey_job_get(self, mock_request):
        result = pyrefinebio.SurveyJob.get(1)
        self.assertObject(result, job_1)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_survey_job_500(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.ServerError):
            pyrefinebio.SurveyJob.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_survey_job_get_404(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.NotFound):
            pyrefinebio.SurveyJob.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_survey_job_search_no_filters(self, mock_request):
        results = pyrefinebio.SurveyJob.search()

        self.assertObject(results[0], job_1)
        self.assertObject(results[1], job_2)

    def test_survey_job_search_with_filters(self):
        filtered_results = pyrefinebio.SurveyJob.search(id=214436, source_type="GEO", success=False)

        for result in filtered_results:
            self.assertEqual(result.id, 214436)
            self.assertEqual(result.source_type, "GEO")
            self.assertFalse(result.success)

    def test_survey_job_search_with_invalid_filters(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidFilters):
            pyrefinebio.SurveyJob.search(foo="bar")



import unittest
from unittest.mock import patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

institutions = [
    {"submitter_institution": "Drexel University"},
    {"submitter_institution": "University of Pennsylvania"},
    {"submitter_institution": "NYU MEDICAL SCHOOL"},
    {"submitter_institution": "Villanova University"},
]


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/institutions/":
        return MockResponse(institutions, url)


class ComputedFileTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computed_file_get(self, mock_request):
        result = pyrefinebio.Institution.search()

        self.assertEqual(len(result), len(institutions))

        for i in range(len(result)):
            self.assertObject(result[i], institutions[i])



import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

qn_target_organisms = [
    {"name": "MUSTELA_PUTORIUS_FURO", "taxonomy_id": 9669},
    {"name": "RATTUS_NORVEGICUS", "taxonomy_id": 10116},
    {"name": "ARABIDOPSIS_THALIANA", "taxonomy_id": 3702},
    {"name": "MUSCULUS", "taxonomy_id": 112137},
]

qn_target = {
    "id": 5741412,
    "filename": "1574800731_target.tsv",
    "size_in_bytes": 318754,
    "is_qn_target": True,
    "sha1": "bb04eb22101c4b25509aae0f2bd0e60763fd053d",
    "s3_bucket": "data-refinery-s3-qn-target-circleci-prod",
    "s3_key": "eo9yywkbvjcc1lj9s0qegnt4_1574800731_target.tsv",
    "s3_url": "https://s3.amazonaws.com/data-refinery-s3-qn-target-circleci-prod/eo9yywkbvjcc1lj9s0qegnt4_1574800731_target.tsv",
    "created_at": "2019-11-26T20:39:29.309489Z",
    "last_modified": "2019-11-26T20:39:29.610226Z",
    "result": {
        "id": 2722294,
        "commands": ["q n _ r e f e r e n c e . p y"],
        "processor": {
            "id": 527,
            "name": "Quantile Normalization Reference",
            "version": "v1.29.6-hotfix",
            "docker_image": "dr_smasher",
            "environment": {
                "R": {"preprocessCore": "1.42.0"},
                "os_pkg": {
                    "r-base": "r-base",
                    "python3": "3.5.1-3",
                    "python3-pip": "8.1.1-2ubuntu0.4",
                },
                "python": {
                    "rpy2": "2.9.5",
                    "Django": "2.2.7",
                    "pandas": "0.23.4",
                    "data-refinery-common": "=v1.29.6-hotfix",
                },
                "os_distribution": "Ubuntu 16.04.6 LTS",
            },
        },
        "is_ccdl": True,
        "annotations": [
            {
                "id": 2790940,
                "data": {
                    "is_qn": True,
                    "geneset": "['ENSCAFG00000025287', 'ENSCAFG00000008570', 'ENSCAFG00000018001']",
                    "samples": ["GSM1540469", "GSM297356", "GSM879434", "GSM879435", "GSM879436"],
                    "organism_id": 28,
                    "num_valid_inputs": 146,
                    "platform_accession_code": "canine2",
                },
                "is_ccdl": True,
                "created_at": "2019-11-26T20:39:29.317850Z",
                "last_modified": "2019-11-26T20:39:29.317850Z",
            }
        ],
        "organism_index": None,
        "time_start": "2019-11-26T20:38:51.632946Z",
        "time_end": "2019-11-26T20:39:28.534756Z",
        "created_at": "2019-11-26T20:39:29.303846Z",
        "last_modified": "2019-11-26T20:39:29.303846Z",
    },
}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/qn_targets/MUSTELA_PUTORIUS_FURO/":
        return MockResponse(qn_target, url)

    if url == "https://api.refine.bio/v1/qn_targets/HUMAN/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/qn_targets/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/qn_targets/":
        return MockResponse(qn_target_organisms, url)


class QNTargetTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_qn_target_get(self, mock_request):
        result = pyrefinebio.QNTarget.get("MUSTELA_PUTORIUS_FURO")
        self.assertObject(result, qn_target)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_qn_target_500(self, mock_request):
        with self.assertRaises(Exception):
            pyrefinebio.QNTarget.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_qn_target_404(self, mock_request):
        with self.assertRaises(Exception):
            pyrefinebio.QNTarget.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_qn_target_search(self, mock_request):
        result = pyrefinebio.QNTarget.search()

        self.assertEqual(len(result), len(qn_target_organisms))

        for i in range(len(result)):
            self.assertObject(result[i], qn_target_organisms[i])



import unittest
from unittest.mock import patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

computational_result_1 = {
    "id": 1,
    "commands": ["test1", "test2"],
    "processor": {
        "id": 0,
        "name": "test_processor",
        "version": "1",
        "docker_image": "test_image",
        "environment": {},
    },
    "is_ccdl": True,
    "annotations": [
        {
            "id": 0,
            "data": "annotation",
            "is_ccdl": True,
            "created_at": "2020-10-02T18:48:24Z",
            "last_modified": "2020-10-02T18:48:24Z",
        }
    ],
    "files": [
        {
            "id": 0,
            "filename": "string",
            "size_in_bytes": -9223372036854776000,
            "is_smashable": True,
            "is_qc": True,
            "sha1": "string",
            "s3_bucket": "string",
            "s3_key": "string",
            "created_at": "2020-10-02T18:48:24Z",
            "last_modified": "2020-10-02T18:48:24Z",
        }
    ],
    "organism_index": {
        "id": 0,
        "assembly_name": "string",
        "organism_name": "string",
        "database_name": "string",
        "release_version": "string",
        "index_type": "string",
        "salmon_version": "string",
        "download_url": "string",
        "result_id": "string",
        "last_modified": "2020-10-02T18:48:24Z",
    },
    "time_start": "2020-10-02T18:48:24Z",
    "time_end": "2020-10-02T18:48:24Z",
    "created_at": "2020-10-02T18:48:24Z",
    "last_modified": "2020-10-02T18:48:24Z",
}

computational_result_2 = {
    "id": 2,
    "commands": ["test1", "test2"],
    "processor": {
        "id": 0,
        "name": "test_processor",
        "version": "1",
        "docker_image": "test_image",
        "environment": {},
    },
    "is_ccdl": True,
    "annotations": [
        {
            "id": 0,
            "data": "annotation",
            "is_ccdl": True,
            "created_at": "2020-10-02T18:48:24Z",
            "last_modified": "2020-10-02T18:48:24Z",
        }
    ],
    "files": [
        {
            "id": 0,
            "filename": "string",
            "size_in_bytes": -9223372036854776000,
            "is_smashable": True,
            "is_qc": True,
            "sha1": "string",
            "s3_bucket": "string",
            "s3_key": "string",
            "created_at": "2020-10-02T18:48:24Z",
            "last_modified": "2020-10-02T18:48:24Z",
        }
    ],
    "organism_index": {
        "id": 0,
        "assembly_name": "string",
        "organism_name": "string",
        "database_name": "string",
        "release_version": "string",
        "index_type": "string",
        "salmon_version": "string",
        "download_url": "string",
        "result_id": "string",
        "last_modified": "2020-10-02T18:48:24Z",
    },
    "time_start": "2020-10-02T18:48:24Z",
    "time_end": "2020-10-02T18:48:24Z",
    "created_at": "2020-10-02T18:48:24Z",
    "last_modified": "2020-10-02T18:48:24Z",
}

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [computational_result_1]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [computational_result_2]}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/computational_results/1/":
        return MockResponse(computational_result_1, url)

    if url == "https://api.refine.bio/v1/computational_results/2/":
        return MockResponse(computational_result_2, url)

    if url == "https://api.refine.bio/v1/computational_results/0/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/computational_results/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/computational_results/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)


class ComputationalResultTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computational_result_get(self, mock_request):
        result = pyrefinebio.ComputationalResult.get(1)
        self.assertObject(result, computational_result_1)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computational_result_500(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.ServerError):
            pyrefinebio.ComputationalResult.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computational_result_get_404(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.NotFound):
            pyrefinebio.ComputationalResult.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computational_result_search_no_filters(self, mock_request):
        results = pyrefinebio.ComputationalResult.search()

        self.assertObject(results[0], computational_result_1)
        self.assertObject(results[1], computational_result_2)

    def test_computational_result_search_with_filters(self):
        filtered_results = pyrefinebio.ComputationalResult.search(processor__id=12)

        for result in filtered_results:
            self.assertEqual(result.processor.id, 12)

    def test_computational_result_search_with_invalid_filters(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidFilters):
            pyrefinebio.ComputationalResult.search(foo="bar")



from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.util import create_paginated_list


class Organism(Base):
    """Organism.

    Retrieve an Organism by name

        >>> import pyrefinebio
        >>> name = "GORILLA"
        >>> organism = pyrefinebio.Organism.get(name)

    Retrieve a list of Organisms

        >>> import pyrefinebio
        >>> organisms = pyrefinebio.Organism.search(has_compendia=True)
    """

    def __init__(
        self, name=None, taxonomy_id=None, has_compendia=None, has_quantfile_compendia=None
    ):
        super().__init__(identifier=name)

        self.name = name
        self.taxonomy_id = taxonomy_id
        self.has_compendia = has_compendia
        self.has_quantfile_compendia = has_quantfile_compendia

    @classmethod
    def get(cls, name):
        """Retrieve an Organism based on name

        Returns:
            Organism

        Parameters:
            name (str): the name for the Organism you want to get
        """

        response = get_by_endpoint("organisms/" + name).json()
        return Organism(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of Organisms based on filters

        Returns:
            list of Organism

        Keyword Arguments:
            has_compendia (bool): filter based on if this Organism has a normalized Compendium
                                  associated with it

            has_quantfile_compendia (bool): filter based on if this Organism has an
                                            RNA-seq Sample Compendium associated with it
        """
        response = get_by_endpoint("organisms", params=kwargs)
        return create_paginated_list(cls, response)



from pyrefinebio import organism as prb_organism
from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.util import parse_date


class QNTarget(Base):
    """QN Target.

    Retrieve a QN target by organism name

        >>> import pyrefinebio
        >>> organism_name = "GORILLA"
        >>> qn_target = pyrefinebio.QNTarget.get(organism_name)

    Retrieve a list of QN target organisms

        >>> import pyrefinebio
        >>> qn_target_organisms = pyrefinebio.QNTarget.search()
    """

    def __init__(
        self,
        id=None,
        filename=None,
        size_in_bytes=None,
        is_qn_target=None,
        sha1=None,
        s3_bucket=None,
        s3_key=None,
        s3_url=None,
        created_at=None,
        last_modified=None,
        result=None,
    ):
        super().__init__(identifier=id)

        self.id = id
        self.filename = filename
        self.size_in_bytes = size_in_bytes
        self.is_qn_target = is_qn_target
        self.sha1 = sha1
        self.s3_bucket = s3_bucket
        self.s3_key = s3_key
        self.s3_url = s3_url
        self.created_at = parse_date(created_at)
        self.last_modified = parse_date(last_modified)
        self.result = result

    @classmethod
    def get(cls, organism_name):
        """Retrieve a QNTarget based on organism name

        Returns:
            QNTarget

        Parameters:
            organism_name (str): the name of the organism for the QNTarget you want to get
        """
        response = get_by_endpoint("qn_targets/" + organism_name).json()
        return QNTarget(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of Organisms that have available QNTargets

        Returns:
            list of Organism

        Since there are no filters, this method always returns all Organisms that have available QNTargets
        """
        response = get_by_endpoint("qn_targets", params=kwargs).json()
        return [prb_organism.Organism(**qn_organism) for qn_organism in response]



import unittest
from copy import deepcopy
from unittest.mock import patch

import pyrefinebio
from pyrefinebio import computed_file as prb_computed_file, original_file as prb_original_file
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse
from tests.test_experiment import experiment_search_result_1, experiment_search_result_2

sample_1 = {
    "id": 1159147,
    "title": "vESRD_CV22",
    "accession_code": "SRR5445147",
    "source_database": "SRA",
    "organism": {"name": "HOMO_SAPIENS", "taxonomy_id": 9606},
    "platform_accession_code": "IonTorrentProton",
    "platform_name": "Ion Torrent Proton",
    "pretty_platform": "Ion Torrent Proton (IonTorrentProton)",
    "technology": "RNA-SEQ",
    "manufacturer": "ION_TORRENT",
    "protocol_info": [
        {
            "Reference": "https://www.ebi.ac.uk/ena/data/view/SRP103842",
            "Description": "EDTA-treated whole blood samples from individual study subjects were centrifuged at 1,600 g for 10 min at 4 °C, transferred to a new centrifuge tube and centrifuged again at 16,000 g for 10 min to remove any residual blood cells. Plasma samples were stored at -80 °C until the step of RNA extraction. Six ml Trizol LS was mixed with 2 ml plasma sample for 5 min at room temperature (RT), followed by mixing with 1 ml 1-bromo-3-chloropropane, vortexed and incubated at 4 °C for 10 min, centrifuged at 6,000 g for 30 min for phase separation. The aqueous phase layer (~ 4 ml) was transferred into a new 50 ml centrifuge tube, mixed with 40 ul glycogen and 5 ml isopropanol, incubated at -20°C for 15 min, centrifuged at 6,000 g for 30 min at 4 °C. After removal of the supernatant, the pellet was lysed with the mixture of Lysis Solution (300 ml) and miRNA Homogenate Additive (30ml) from mirVana (Thermo Fisher, AM1650) kit, vortexed and fully resuspended, followed by addition of 350 ml isopropanol. The mixture was transferred onto an RNAqueous micro filter cartridge and centrifuged at 10,000 g for 15 sec, washed with 700 ml Wash Solution 1 and centrifuged at 10,000 g for 10 sec, washed again with 500 ml Wash Solution 2/3 and centrifuged at 10,000 g for 1 min. Eight ml of Elution Solution (preheated to 65°C) was applied to the filter and centrifuged 10 sec at 10,000 g to obtain RNA samples. The final step was repeated again and the RNA samples were mixed, quantified using Qubit RNA HS assay kit and stored at -80 °C until RNA sequencing. (Exploratory cohort) For the plasma transcriptome RNASeq in the exploratory cohort, 10 ng of total plasma RNA samples from individual study subjects was reversed transcribed to cDNA followed by PCR amplification (16 cycles), digestion with FuPa enzyme, ligation (60 minutes) and Ampure cleanup according to the user guide supplied with the Ion AmpliSeq Transcriptome Human Gene Expression Kit (Thermo Fisher Scientific, PN A26325).  Barcoded libraries were quantitated by Ion Library TaqMan™ Quantitation Kit (Thermo Fisher Scientific, PN4468802) and combined to either 7- or 8- plex pools for sequencing. Sequencing template preparation and Proton P1 chip loading were performed on the Ion Chef™ instrument using the Ion Proton™ Hi-Q ™ Chef kit. (Validation cohort) For the plasma transcriptome RNASeq in the validation cohort, 10 ng of total plasma RNA samples from individual study subjects was reversed transcribed to cDNA followed by PCR amplification (16 cycles) using predesigned primers targeting 8 candidate lncRNAs and 6 control mRNAs, digestion with FuPa enzyme, ligation (60 minutes) and Ampure cleanup.  Barcoded libraries were quantitated by Ion Library TaqMan™ Quantitation Kit (Thermo Fisher Scientific, PN4468802) and pooled for sequencing. Sequencing template preparation and Proton P1 chip loading were performed on the Ion Chef™ instrument using the Ion Proton™ Hi-Q ™ Chef kit.",
        }
    ],
    "annotations": [
        {
            "data": {
                "type": ["RNA"],
                "title": ["sample_150"],
                "status": ["Public on Jan 20 2016"],
            },
            "is_ccdl": False,
            "created_at": "2018-12-19T20:15:49.129080Z",
            "last_modified": "2018-12-19T20:15:49.129080Z",
        }
    ],
    "results": [
        {
            "id": 1339591,
            "processor": {
                "id": 302,
                "name": "Salmon Quant",
                "version": "v1.21.2-hotfix",
                "docker_image": "dr_salmon",
                "environment": {
                    "os_pkg": {"python3": "3.5.1-3", "python3-pip": "8.1.1-2ubuntu0.4"},
                    "python": {"Django": "2.1.8", "data-refinery-common": "=v1.21.2-hotfix"},
                    "cmd_line": {
                        "salmon --version": "salmon 0.13.1",
                        "sra-stat --version": "sra-stat : 2.9.1",
                    },
                    "os_distribution": "Ubuntu 16.04.5 LTS",
                },
            },
            "organism_index": {
                "id": 539,
                "assembly_name": "GRCh38",
                "organism_name": "HOMO_SAPIENS",
                "database_name": "EnsemblMain",
                "release_version": "96",
                "index_type": "TRANSCRIPTOME_LONG",
                "salmon_version": "salmon 0.13.1",
                "download_url": "https://s3.amazonaws.com/data-refinery-s3-transcriptome-index-circleci-prod/HOMO_SAPIENS_TRANSCRIPTOME_LONG_1559763336.tar.gz",
                "result_id": 1242756,
                "last_modified": "2019-06-05T19:36:09.993557Z",
            },
        },
        {
            "id": 1339595,
            "processor": {
                "id": 303,
                "name": "Salmontools",
                "version": "v1.21.2-hotfix",
                "docker_image": "dr_salmon",
                "environment": {
                    "os_pkg": {
                        "g++": "4:5.3.1-1ubuntu1",
                        "cmake": "3.5.1-1ubuntu3",
                        "python3": "3.5.1-3",
                        "python3-pip": "8.1.1-2ubuntu0.4",
                    },
                    "python": {"Django": "2.1.8", "data-refinery-common": "=v1.21.2-hotfix"},
                    "cmd_line": {"salmontools --version": "Salmon Tools 0.1.0"},
                    "os_distribution": "Ubuntu 16.04.5 LTS",
                },
            },
            "organism_index": None,
        },
    ],
    "source_archive_url": "",
    "has_raw": True,
    "sex": "",
    "age": None,
    "specimen_part": "",
    "genetic_information": "",
    "developmental_stage": "",
    "disease": "",
    "disease_stage": "",
    "cell_line": "",
    "treatment": "",
    "race": "",
    "subject": "human plasma",
    "compound": "",
    "time": "",
    "is_processed": True,
    "created_at": "2019-07-13T14:05:37.129910Z",
    "last_modified": "2019-07-14T16:29:34.433025Z",
    "original_files": [1715994],
    "computed_files": [2425581, 2427322, 2427387],
    "experiment_accession_codes": ["SRP103842"],
}

sample_2 = {
    "id": 840023,
    "title": "sample_150",
    "accession_code": "GSM1562009",
    "source_database": "GEO",
    "organism": {"name": "HOMO_SAPIENS", "taxonomy_id": 9606},
    "platform_accession_code": "hgu133a2",
    "platform_name": "[HG-U133A_2] Affymetrix Human Genome U133A 2.0 Array",
    "pretty_platform": "Affymetrix Human Genome U133A 2.0 Array (hgu133a2)",
    "technology": "MICROARRAY",
    "manufacturer": "AFFYMETRIX",
    "protocol_info": {
        "Reference": "https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSM1562009",
        "Scan protocol": ["Affymetrix GeneChip Scanner"],
        "Label protocol": [
            "Total RNA is converted into double stranded cDNA via reverse transcription (RT) using an oligo-d(T) primer-adaptor. During this RT step, four peptide nucleic acid (PNA) oligomers that bind portions of the alpha and beta hemoglobin transcripts are added to prevent transcription of globin mRNAs.  The cDNA is then converted to dsDNA using T4 DNA polymerase.  The resulting dsDNA is purified to remove unincorporated nucleotides, salts, enzymes and inorganic phosphates.  This is used as a template for in vitro transcription using T7 RNA polymerase and biotinylated ribonucleotides. The resulting cRNA is purified and quantitated using spectrophotometry.  Globin reduction in the cRNA is verified using an Agilent Bioanalyzer."
        ],
        "Data processing": ["Expression Console"],
        "Extraction protocol": [
            "RNA isolation was performed from blood collected in PAXgene Blood RNA Tubes using the PAXgene 96 Blood RNA Kit (QIAgen).  Briefly, blood was pelleted, washed, resuspended, and incubated in optimized buffers with Proteinase K.  The lysate was applied to a filter plate and centrifuged to remove any residual cell debris.  Ethanol was added to the flow-through to provide appropriate binding conditions for the RNA.  The lysate was then applied to an RNA plate where the RNA selectively binds to membrane.  Residual DNA was removed through a DNase digestion on the membrane and any remaining contaminants were washed away.  High-quality RNA was then eluted in nuclease-free water."
        ],
        "Hybridization protocol": [
            "Next, 11ug of purified cRNA is fragmented using a 5X fragmentation buffer, then a hybridization cocktail is prepared and added to the fragmentation product using the Hybridization, Wash and Stain kit (Affymetrix), applied to arrays, and incubated at 45C for 16 hours."
        ],
    },
    "annotations": [
        {
            "data": {
                "type": ["RNA"],
                "title": ["sample_150"],
                "status": ["Public on Jan 20 2016"],
                "label_ch1": ["biotin"],
            },
            "is_ccdl": False,
            "created_at": "2018-12-19T20:15:49.129080Z",
            "last_modified": "2018-12-19T20:15:49.129080Z",
        }
    ],
    "results": [
        {
            "id": 1339595,
            "processor": {
                "id": 303,
                "name": "Salmontools",
                "version": "v1.21.2-hotfix",
                "docker_image": "dr_salmon",
                "environment": {
                    "os_pkg": {
                        "g++": "4:5.3.1-1ubuntu1",
                        "cmake": "3.5.1-1ubuntu3",
                        "python3": "3.5.1-3",
                        "python3-pip": "8.1.1-2ubuntu0.4",
                    },
                    "python": {"Django": "2.1.8", "data-refinery-common": "=v1.21.2-hotfix"},
                    "cmd_line": {"salmontools --version": "Salmon Tools 0.1.0"},
                    "os_distribution": "Ubuntu 16.04.5 LTS",
                },
            },
            "organism_index": None,
        }
    ],
    "source_archive_url": "",
    "has_raw": True,
    "sex": "",
    "age": None,
    "specimen_part": "",
    "genetic_information": "",
    "developmental_stage": "",
    "disease": "",
    "disease_stage": "",
    "cell_line": "",
    "treatment": "",
    "race": "",
    "subject": "",
    "compound": "",
    "time": "",
    "is_processed": True,
    "created_at": "2018-12-19T20:15:46.399775Z",
    "last_modified": "2019-08-09T20:44:29.963746Z",
    "original_files": [1920272, 1208147],
    "computed_files": [2679162],
    "experiment_accession_codes": ["GSE63990"],
}

sample_1_object_dict = deepcopy(sample_1)

sample_1_object_dict["original_files"] = [
    prb_original_file.OriginalFile(file_id) for file_id in sample_1["original_files"]
]
sample_1_object_dict["computed_files"] = [
    prb_computed_file.ComputedFile(file_id) for file_id in sample_1["computed_files"]
]

sample_2_object_dict = deepcopy(sample_2)

sample_2_object_dict["original_files"] = [
    prb_original_file.OriginalFile(file_id) for file_id in sample_2["original_files"]
]
sample_2_object_dict["computed_files"] = [
    prb_computed_file.ComputedFile(file_id) for file_id in sample_2["computed_files"]
]

result = {
    "id": 1339595,
    "commands": [
        "salmontools extract-unmapped -u /home/user/data_store/processor_job_2893940/SRR5445147_output/aux_info/unmapped_names.txt -o /home/user/data_store/processor_job_2893940/salmontools/unmapped_by_salmon -r /home/user/data_store/processor_job_2893940/SRR5445147"
    ],
    "processor": {
        "id": 303,
        "name": "Salmontools",
        "version": "v1.21.2-hotfix",
        "docker_image": "dr_salmon",
        "environment": {
            "os_pkg": {
                "g++": "4:5.3.1-1ubuntu1",
                "cmake": "3.5.1-1ubuntu3",
                "python3": "3.5.1-3",
                "python3-pip": "8.1.1-2ubuntu0.4",
            },
            "python": {"Django": "2.1.8", "data-refinery-common": "=v1.21.2-hotfix"},
            "cmd_line": {"salmontools --version": "Salmon Tools 0.1.0"},
            "os_distribution": "Ubuntu 16.04.5 LTS",
        },
    },
    "is_ccdl": True,
    "annotations": [],
    "files": [
        {
            "id": 2425581,
            "filename": "salmontools-result.tar.gz",
            "size_in_bytes": 190,
            "is_smashable": False,
            "is_qc": True,
            "sha1": "9d3396dffe4617f6d2d8414ebb2a9f132a0f3c24",
            "s3_bucket": "data-refinery-s3-circleci-prod",
            "s3_key": "ysccy3qawpt2awi1hm4alaz6_salmontools-result.tar.gz",
            "created_at": "2019-07-14T14:49:50.224352Z",
            "last_modified": "2019-07-14T14:49:50.493387Z",
        }
    ],
    "organism_index": None,
    "time_start": "2019-07-14T14:49:47.567197Z",
    "time_end": "2019-07-14T14:49:49.418108Z",
    "created_at": "2019-07-14T14:49:50.210938Z",
    "last_modified": "2019-07-14T14:49:50.210938Z",
}

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [sample_1]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [sample_2]}

e_search_1 = {
    "count": 2,
    "next": "e_search_2",
    "previous": None,
    "results": [experiment_search_result_1],
}

e_search_2 = {
    "count": 2,
    "next": None,
    "previous": "e_search_1",
    "results": [experiment_search_result_2],
}


def mock_request(method, url, **kwargs):
    if url == "https://api.refine.bio/v1/samples/SRR5445147/":
        return MockResponse(sample_1, url)

    if url == "https://api.refine.bio/v1/samples/GSM1562009/":
        return MockResponse(sample_2, url)

    # this request needs to exist because the result's organism index is None for both test samples
    # so a request will be made to make sure that that value should actually be None
    if url == "https://api.refine.bio/v1/computational_results/1339595/":
        return MockResponse(result, url)

    if url == "https://api.refine.bio/v1/samples/0/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/samples/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/samples/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)

    if url == "https://api.refine.bio/v1/search/":
        return MockResponse(e_search_1, "e_search_2")

    if url == "e_search_2":
        return MockResponse(e_search_2, url)


class SampleTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_sample_get(self, mock_request):
        result = pyrefinebio.Sample.get("SRR5445147")
        self.assertObject(result, sample_1_object_dict)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_sample_500(self, mock_request):
        with self.assertRaises(Exception):
            pyrefinebio.Sample.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_sample_404(self, mock_request):
        with self.assertRaises(Exception):
            pyrefinebio.Sample.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_sample_result_is_fully_populated(self, mock_request):
        result = pyrefinebio.Sample.get("SRR5445147")
        self.assertIsNotNone(result.results[1].is_ccdl)
        self.assertIsNotNone(result.results[1].commands)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_sample_search_no_filters(self, mock_request):
        results = pyrefinebio.Sample.search()

        self.assertObject(results[0], sample_1_object_dict)
        self.assertObject(results[1], sample_2_object_dict)

    def test_sample_search_with_filters(self):
        filtered_results = pyrefinebio.Sample.search(
            limit=1, organism=258, has_raw=True, is_processed=False
        )

        self.assertGreater(len(filtered_results), 0)
        for result in filtered_results:
            self.assertEqual(result.organism.name, "MUS")
            self.assertTrue(result.has_raw)
            self.assertFalse(result.is_processed)

    def test_sample_search_with_invalid_filters(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidFilters):
            pyrefinebio.Sample.search(foo="bar")

    def test_sample_search_with_invalid_filter_type(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidFilterType):
            pyrefinebio.Sample.search(age="foo")

    def test_sample_search_with_multiple_invalid_filter_types(self):
        with self.assertRaises(pyrefinebio.exceptions.MultipleErrors):
            pyrefinebio.Sample.search(age="foo", organism="bar")

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_sample_get_experiments(self, mock_request):
        result = pyrefinebio.Sample.get("SRR5445147")
        experiments = list(result.experiments)

        self.assertObject(experiments[0], experiment_search_result_1)
        self.assertObject(experiments[1], experiment_search_result_2)



import unittest
from copy import deepcopy
from unittest.mock import patch

import pyrefinebio
from pyrefinebio import original_file as prb_original_file
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

job_1 = {
    "id": 1,
    "pipeline_applied": "JANITOR",
    "num_retries": 0,
    "retried": False,
    "worker_id": None,
    "ram_amount": 2048,
    "volume_index": "-1",
    "worker_version": None,
    "failure_reason": None,
    "batch_job_id": None,
    "batch_job_queue": None,
    "success": None,
    "original_files": [],
    "datasets": [],
    "start_time": None,
    "end_time": None,
    "created_at": "2020-10-14T17:30:09.480064Z",
    "last_modified": "2020-10-14T17:30:09.493173Z",
}

job_2 = {
    "id": 2,
    "pipeline_applied": "SMASHER",
    "num_retries": 0,
    "retried": False,
    "worker_id": "i-01877387b773bc7b4",
    "ram_amount": 4096,
    "volume_index": None,
    "worker_version": "v1.39.11",
    "failure_reason": None,
    "batch_job_id": "faf15c0e-3164-4bae-98f6-c6668f9ff4a5",
    "batch_job_queue": "data-refinery-batch-workers-queue-circleci-prod-0",
    "success": True,
    "original_files": [12345, 13456, 14567],
    "datasets": ["test-set-1", "test-set-2"],
    "start_time": "2020-10-14T17:00:38.579600Z",
    "end_time": "2020-10-14T17:02:12.439331Z",
    "created_at": "2020-10-14T17:00:04.154755Z",
    "last_modified": "2020-10-14T17:02:12.658330Z",
}

job_1_object_dict = deepcopy(job_1)

job_1_object_dict["original_files"] = [
    prb_original_file.OriginalFile(file_id) for file_id in job_1["original_files"]
]

job_2_object_dict = deepcopy(job_2)

job_2_object_dict["original_files"] = [
    prb_original_file.OriginalFile(file_id) for file_id in job_2["original_files"]
]

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [job_1]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [job_2]}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/jobs/processor/1/":
        return MockResponse(job_1, url)

    if url == "https://api.refine.bio/v1/jobs/processor/2/":
        return MockResponse(job_2, url)

    if url == "https://api.refine.bio/v1/jobs/processor/0/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/jobs/processor/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/jobs/processor/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)


class ProcessorJobTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_processor_job_get(self, mock_request):
        result = pyrefinebio.ProcessorJob.get(1)
        self.assertObject(result, job_1_object_dict)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_processor_job_500(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.ServerError):
            pyrefinebio.ProcessorJob.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_processor_job_get_404(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.NotFound):
            pyrefinebio.ProcessorJob.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_processor_job_search_no_filters(self, mock_request):
        results = pyrefinebio.ProcessorJob.search()

        self.assertObject(results[0], job_1_object_dict)
        self.assertObject(results[1], job_2_object_dict)

    def test_processor_job_search_with_filters(self):
        filtered_results = pyrefinebio.ProcessorJob.search(
            retried=True, num_retries=4, pipeline_applied="QN_REFERENCE"
        )

        for result in filtered_results:
            self.assertTrue(result.retried)
            self.assertEqual(result.num_retries, 4)
            self.assertEqual(result.pipeline_applied, "QN_REFERENCE")

    def test_processor_job_search_with_invalid_filters(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidFilters):
            pyrefinebio.ProcessorJob.search(foo="bar")



import unittest
from copy import deepcopy
from unittest.mock import patch

import pyrefinebio
from pyrefinebio import job as prb_job, original_file as prb_original_file
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

processor_job = {
    "id": 29708302,
    "pipeline_applied": "SALMON",
    "num_retries": 0,
    "retried": False,
    "worker_id": "i-0ea363205be30776f",
    "ram_amount": 12288,
    "volume_index": "i-0ea363205be30776f",
    "worker_version": "v1.39.5",
    "failure_reason": None,
    "batch_job_id": "2d4174c9-c449-49c5-b9c2-f270ec4597f0",
    "batch_job_queue": "data-refinery-batch-workers-queue-circleci-prod-0",
    "success": True,
    "original_files": [2925811],
    "datasets": [],
    "start_time": "2020-10-01T20:20:04.109061Z",
    "end_time": "2020-10-01T20:23:43.507988Z",
    "created_at": "2020-10-01T20:19:05.841594Z",
    "last_modified": "2020-10-01T20:23:43.508066Z",
}

processor_job_object_dict = deepcopy(processor_job)

processor_job_object_dict["original_files"] = [
    prb_original_file.OriginalFile(file_id) for file_id in processor_job["original_files"]
]

downloader_job = {
    "id": 7381811,
    "downloader_task": "SRA",
    "num_retries": 0,
    "retried": False,
    "was_recreated": False,
    "worker_id": "i-0ea363205be30776f",
    "worker_version": "v1.39.5",
    "batch_job_id": "faf15c0e-3164-4bae-98f6-c6668f9ff4a5",
    "batch_job_queue": "data-refinery-batch-workers-queue-circleci-prod-0",
    "failure_reason": None,
    "success": True,
    "original_files": [2925811],
    "start_time": "2020-10-01T20:18:23.079796Z",
    "end_time": "2020-10-01T20:19:05.873400Z",
    "created_at": "2020-10-01T20:14:59.668642Z",
    "last_modified": "2020-10-01T20:19:05.873426Z",
}

downloader_job_object_dict = deepcopy(downloader_job)

downloader_job_object_dict["original_files"] = [
    prb_original_file.OriginalFile(file_id) for file_id in downloader_job["original_files"]
]

og_file_1 = {
    "id": 1,
    "filename": "SRR067396.sra",
    "size_in_bytes": 1004703052,
    "sha1": "cdef439e998b12e5f58a89b92f5a96d01aecbe75",
    "samples": [
        {
            "accession_code": "SRR067396",
            "platform_name": "Illumina Genome Analyzer II",
            "pretty_platform": "Illumina Genome Analyzer II (IlluminaGenomeAnalyzerII)",
            "technology": "RNA-SEQ",
            "is_processed": True,
        }
    ],
    "processor_jobs": [processor_job_object_dict],
    "downloader_jobs": [downloader_job_object_dict],
    "source_url": "anonftp@ftp-private.ncbi.nlm.nih.gov:/sra/sra-instant/reads/ByRun/sra/SRR/SRR067/SRR067396/SRR067396.sra",
    "source_filename": "SRR067396.sra",
    "is_downloaded": False,
    "is_archive": False,
    "has_raw": True,
    "created_at": "2020-10-01T20:13:03.302659Z",
    "last_modified": "2020-10-01T20:23:43.278755Z",
}

og_file_list_1 = {
    "id": 1,
    "filename": "SRR067396.sra",
    "size_in_bytes": 1004703052,
    "sha1": "cdef439e998b12e5f58a89b92f5a96d01aecbe75",
    "samples": [123456],
    "processor_jobs": [29708302],
    "downloader_jobs": [7381811],
    "source_url": "anonftp@ftp-private.ncbi.nlm.nih.gov:/sra/sra-instant/reads/ByRun/sra/SRR/SRR067/SRR067396/SRR067396.sra",
    "source_filename": "SRR067396.sra",
    "is_archive": False,
    "has_raw": True,
    "created_at": "2020-10-01T20:13:03.302659Z",
    "last_modified": "2020-10-01T20:23:43.278755Z",
}

og_file_list_2 = {
    "id": 4,
    "filename": "GSM2351172_A52084400939855091415421246947329.CEL",
    "samples": [123456, 224564],
    "size_in_bytes": 69357911,
    "sha1": "9a7068160ae899de6bb245f5e35982a4810cb36e",
    "processor_jobs": [123456, 222111, 444333],
    "downloader_jobs": [123456, 123457, 123458],
    "source_url": "ftp://ftp.ncbi.nlm.nih.gov/geo/samples/GSM2351nnn/GSM2351172/suppl/GSM2351172_A52084400939855091415421246947329.CEL.gz",
    "is_archive": True,
    "source_filename": "GSM2351172_A52084400939855091415421246947329.CEL.gz",
    "has_raw": True,
    "created_at": "2020-10-13T23:03:56.233937Z",
    "last_modified": "2020-10-13T23:03:56.233937Z",
}

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [og_file_list_1]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [og_file_list_2]}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/original_files/1/":
        return MockResponse(og_file_1, url)

    if url == "https://api.refine.bio/v1/original_files/0/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/original_files/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/jobs/processor/29708302/":
        return MockResponse(processor_job, url)

    if url == "https://api.refine.bio/v1/jobs/downloader/7381811/":
        return MockResponse(downloader_job, url)

    if url == "https://api.refine.bio/v1/original_files/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)


class OriginalFileTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_original_file_get(self, mock_request):
        result = pyrefinebio.OriginalFile.get(1)
        self.assertObject(result, og_file_1)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_original_file_500(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.ServerError):
            pyrefinebio.OriginalFile.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_original_file_get_404(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.NotFound):
            pyrefinebio.OriginalFile.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_original_file_search_no_filters(self, mock_request):
        results = pyrefinebio.OriginalFile.search()

        self.assertObject(results[0], og_file_list_1)
        self.assertObject(results[1], og_file_list_2)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_original_file_search_fully_populated(self, mock_request):
        # for some reason /original_files search endpoint doesn't return is_downloaded
        # make sure this property is populated correctly
        result = pyrefinebio.OriginalFile.search()[0]
        self.assertIsNotNone(result.is_downloaded)

    def test_original_file_search_with_filters(self):
        filtered_results = pyrefinebio.OriginalFile.search(size_in_bytes=1000)

        for result in filtered_results:
            self.assertEqual(result.size_in_bytes, 1000)

    def test_original_file_search_with_invalid_filters(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidFilters):
            pyrefinebio.OriginalFile.search(foo="bar")



from pyrefinebio import (
    annotation as prb_annotation,
    computational_result as prb_computational_result,
    computed_file as prb_computed_file,
    experiment as prb_experiment,
    organism as prb_organism,
    original_file as prb_original_file,
    job as prb_job,
)
from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.util import create_paginated_list, parse_date


class Sample(Base):
    """Sample.

    Retrieve a Sample based on accession code

        >>> import pyrefinebio
        >>> accession_code = "GSM000000"
        >>> sample = pyrefinebio.Sample.get(accession_code)

    Retrieve a list of Samples based on filters

        >>> import pyrefinebio
        >>> samples = pyrefinebio.Sample.search(is_processed=True, specimen_part="soft-tissue sarcoma")
    """

    def __init__(
        self=None,
        id=None,
        title=None,
        accession_code=None,
        source_database=None,
        organism=None,
        platform_accession_code=None,
        platform_name=None,
        pretty_platform=None,
        technology=None,
        manufacturer=None,
        protocol_info=None,
        annotations=None,
        results=None,
        source_archive_url=None,
        has_raw=None,
        sex=None,
        age=None,
        specimen_part=None,
        genetic_information=None,
        developmental_stage=None,
        disease=None,
        disease_stage=None,
        cell_line=None,
        treatment=None,
        race=None,
        subject=None,
        compound=None,
        time=None,
        is_processed=None,
        is_unable_to_be_processed=None,
        created_at=None,
        last_modified=None,
        contributed_metadata=None,
        contributed_keywords=None,
        original_files=[],
        computed_files=[],
        last_processor_job=None,
        last_downloader_job=None,
        most_recent_smashable_file=None,
        most_recent_quant_file=None,
        experiment_accession_codes=None,
        experiments=None,
    ):
        super().__init__(identifier=accession_code)

        self.id = id
        self.title = title
        self.accession_code = accession_code
        self.source_database = source_database
        self.organism = prb_organism.Organism(**(organism)) if organism else None
        self.platform_accession_code = platform_accession_code
        self.platform_name = platform_name
        self.pretty_platform = pretty_platform
        self.technology = technology
        self.manufacturer = manufacturer
        self.protocol_info = protocol_info
        self.annotations = (
            [prb_annotation.Annotation(**annotation) for annotation in annotations]
            if annotations
            else []
        )
        self.results = (
            [prb_computational_result.ComputationalResult(**result) for result in results]
            if results
            else []
        )
        self.source_archive_url = source_archive_url
        self.has_raw = has_raw
        self.sex = sex
        self.age = age
        self.specimen_part = specimen_part
        self.genetic_information = genetic_information
        self.developmental_stage = developmental_stage
        self.disease = disease
        self.disease_stage = disease_stage
        self.cell_line = cell_line
        self.treatment = treatment
        self.race = race
        self.subject = subject
        self.compound = compound
        self.time = time
        self.is_processed = is_processed
        self.is_unable_to_be_processed = is_unable_to_be_processed
        self.created_at = parse_date(created_at)
        self.last_modified = parse_date(last_modified)
        self.contributed_metadata = contributed_metadata
        self.contributed_keywords = contributed_keywords

        self.original_files = (
            [prb_original_file.OriginalFile(id=file_id) for file_id in original_files]
            if original_files
            else []
        )
        self.computed_files = (
            [prb_computed_file.ComputedFile(id=file_id) for file_id in computed_files]
            if computed_files
            else []
        )

        # this isn't populated yet but the api does include these keys in the response
        # so for now let's just try to apply them
        if last_processor_job:
            self.last_processor_job = prb_job.ProcessorJob(**last_processor_job)
        if last_downloader_job:
            self.last_downloader_job = prb_job.DownloaderJob(**last_downloader_job)

        self.most_recent_smashable_file = most_recent_smashable_file
        self.most_recent_quant_file = most_recent_quant_file
        self.experiment_accession_codes = experiment_accession_codes
        self.experiments = experiments
        
    @property
    def experiments(self):
        if not self._experiments:
            self._experiments = prb_experiment.Experiment.search(
                accession_code=self.experiment_accession_codes
            )

        return self._experiments

    @experiments.setter
    def experiments(self, value):
        self._experiments = value

    @classmethod
    def get(cls, accession_code):
        """Retrieve a Sample based on its accession code.

        Returns:
            Sample

        Parameters:
            accession_code (str): The accession code for the Sample to be retrieved.
        """
        response = get_by_endpoint("samples/" + accession_code).json()
        return cls(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of Samples based on various filters

        Returns:
            list of Sample

        Keyword Arguments:
            ordering (str): which field to use when ordering the results

            title (str): filter based on the Sample's title

            organism__name (str): filter based on the Organism that the Sample was taken from

            organism__taxonomy_id (int): filter based on the Organism that the Sample was taken from

            source_database (str): filter based on the publically available repository
                                   that the Sample was taken from

            source_archive_url (str): filter based on Sample's source url

            has_raw (bool): filter based on if the Sample had raw data available in the source database

            platform_name (str): filter based on the name of the Platform that was used to assay the Sample

            technology (str): filter based on the technology that was used to assay the Sample

            manufacturer (str): filter based on the manufacturer of the technology that was used to assay the Sample

            sex (str): filter based on the sex information provided by the submitter

            age (number): filter based on the age information provided by the submitter

            specimen_part (str): filter based on the part of the specimen reported by the submitter

            genetic_information (str): filter based on the genetic_information of the subject that the Sample was taken from

            developmental_stage (str): filter based on the developmental_stage of the subject that the Sample was taken from

            disease (str): filter based on the disease information provided by the submitter

            disease_stage (str): filter based on the disease stage information provided by the submitter

            cell_line (str): filter based on the cell line used for the sample

            treatment (str): filter based on the treatment information provided by the submitter

            race (str): filter based on the race information provided by the submitter

            subject (str): filter based on the subject identifier information provided by the submitter

            compound (str): filter based on the compound information provided by the submitter

            time (str): filter based on the time information provided by the submitter

            is_processed (bool): filter based on if the Sample has been processed

            limit (int): number of results to return per page.

            offset (int): the initial index from which to return the results.

            dataset_id (str): filter based on the Dataset ids that the Sample has been added to

            experiment_accession_code (str): filter based on the Experiments that are associated
                                             with the Sample

            accession_codes (str): filter based on multiple accession codes at once
        """
        response = get_by_endpoint("samples", params=kwargs)
        return create_paginated_list(cls, response)



import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

index_1 = {
    "id": 1,
    "assembly_name": "0",
    "organism_name": "CYPRINUS_CARPIO",
    "database_name": "EnsemblMain",
    "release_version": "101",
    "index_type": "TRANSCRIPTOME_SHORT",
    "salmon_version": "salmon 0.13.1",
    "download_url": "test-url",
    "result_id": 2784457,
    "last_modified": "2020-10-01T19:11:44.029922Z",
}

index_2 = {
    "id": 2,
    "assembly_name": "0",
    "organism_name": "HUMAN",
    "database_name": "EnsemblMain",
    "release_version": "101",
    "index_type": "TRANSCRIPTOME_LONG",
    "salmon_version": "salmon 0.13.1",
    "download_url": "test-url",
    "result_id": 2784456,
    "last_modified": "2020-10-01T19:11:06.836302Z",
}

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [index_1]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [index_2]}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/transcriptome_indices/1/":
        return MockResponse(index_1, url)

    if url == "https://api.refine.bio/v1/transcriptome_indices/0/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/transcriptome_indices/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/transcriptome_indices/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)


class TranscriptomeIndexTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_transcriptome_index_get(self, mock_request):
        result = pyrefinebio.TranscriptomeIndex.get(1)
        self.assertObject(result, index_1)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_transcriptome_index_500(self, mock_request):
        with self.assertRaises(Exception):
            pyrefinebio.TranscriptomeIndex.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_transcriptome_index_404(self, mock_request):
        with self.assertRaises(Exception):
            pyrefinebio.TranscriptomeIndex.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_transcriptome_index_search_no_filters(self, mock_request):
        results = pyrefinebio.TranscriptomeIndex.search()

        self.assertObject(results[0], index_1)
        self.assertObject(results[1], index_2)

    def test_transcriptome_index_search_with_filters(self):
        filtered_results = pyrefinebio.TranscriptomeIndex.search(
            index_type="TRANSCRIPTOME_SHORT",
            salmon_version="salmon 0.13.1",
            organism__name="CYPRINUS_CARPIO",
        )

        for result in filtered_results:
            self.assertEqual(result.index_type, "TRANSCRIPTOME_SHORT")
            self.assertEqual(result.salmon_version, "salmon 0.13.1")
            self.assertEqual(result.organism_name, "CYPRINUS_CARPIO")



import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

platforms = [
    {"platform_accession_code": "testcode1", "platform_name": "test platform 1"},
    {"platform_accession_code": "testcode2", "platform_name": "test platform 2"},
    {"platform_accession_code": "testcode3", "platform_name": "test platform 3"},
    {"platform_accession_code": "testcode4", "platform_name": "test platform 4"},
]


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/platforms/":
        return MockResponse(platforms, url)


class PlatfromTest(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computed_file_get(self, mock_request):
        result = pyrefinebio.Platform.search()

        self.assertEqual(len(result), len(platforms))

        for i in range(len(result)):
            self.assertObject(result[i], platforms[i])



from datetime import datetime


class CustomAssertions:
    def assertObject(self, actual, expected):
        """Assert that a model is equivalent to its dict counterpart.

        Verifies that all properties have been set correctly
        Verifies that all set properties can be accessed by dot notation
        """
        # if actual and expected are the same type, just compare with ==
        if type(actual) == type(expected):
            if actual != expected:
                raise AssertionError(
                    "actual did not match expected value:\nexpected: {0}\nactual: {1}".format(
                        expected, actual
                    )
                )
            return

        if isinstance(actual, datetime):
            if actual.microsecond:
                actual = datetime.strftime(actual, "%Y-%m-%dT%H:%M:%S.%fZ")
            else:
                actual = datetime.strftime(actual, "%Y-%m-%dT%H:%M:%SZ")
            if actual != expected:
                raise AssertionError(
                    "actual did not match expected value:\nexpected: {0}\nactual: {1}".format(
                        expected, actual
                    )
                )
            return

        # if we get here, actual is an model class object so do the complicated logic
        for key, value in expected.items():
            actual_value = getattr(actual, key)

            if type(value) is list:
                if len(value) != len(actual_value):
                    raise AssertionError(
                        "actual.{0} did not match expected value:\nexpected: {1} - length: {2}\nactual: {3} - length: {4}".format(
                            key, value, len(value), actual_value, len(actual_value)
                        )
                    )

                for i in range(len(value)):
                    self.assertObject(actual_value[i], value[i])

            else:
                self.assertObject(actual_value, value)



import os
import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

computed_file_1 = {
    "id": 1,
    "filename": "salmontools-result.tar.gz",
    "samples": [
        {
            "accession_code": "SRR067392",
            "platform_name": "Illumina Genome Analyzer II",
            "pretty_platform": "Illumina Genome Analyzer II (IlluminaGenomeAnalyzerII)",
            "technology": "RNA-SEQ",
            "is_processed": True,
        }
    ],
    "size_in_bytes": 191,
    "is_qn_target": False,
    "is_smashable": False,
    "is_qc": True,
    "is_compendia": False,
    "quant_sf_only": False,
    "compendium_version": None,
    "sha1": "77ea7dd477fa9a9ec5f33bc91f46c0e224edeb92",
    "s3_bucket": "data-refinery-s3-circleci-prod",
    "s3_key": "gqo8hxhqm3tcz4c2x9sthhw5_salmontools-result.tar.gz",
    "s3_url": "https://s3.amazonaws.com/data-refinery-s3-circleci-prod/gqo8hxhqm3tcz4c2x9sthhw5_salmontools-result.tar.gz",
    "download_url": "test_download_url",
    "created_at": "2020-10-01T20:23:41.661425Z",
    "last_modified": "2020-10-01T20:23:42.378884Z",
    "result": {
        "id": 2784485,
        "commands": [
            "salmontools extract-unmapped -u /home/user/data_store/processor_job_29708301/SRR067392_output/aux_info/unmapped_names.txt -o /home/user/data_store/processor_job_29708301/salmontools/unmapped_by_salmon -r /home/user/data_store/processor_job_29708301/SRR067392.sra"
        ],
        "processor": {
            "id": 607,
            "name": "Salmontools",
            "version": "v1.39.5",
            "docker_image": "dr_salmon",
            "environment": {
                "os_pkg": {
                    "g++": "4:5.3.1-1ubuntu1",
                    "cmake": "3.5.1-1ubuntu3",
                    "python3": "3.5.1-3",
                    "python3-pip": "8.1.1-2ubuntu0.4",
                },
                "python": {"Django": "2.2.13", "data-refinery-common": "1.39.5"},
                "cmd_line": {"salmontools --version": "Salmon Tools 0.1.0"},
                "os_distribution": "Ubuntu 16.04.6 LTS",
            },
        },
        "is_ccdl": True,
        "annotations": [],
        "files": [
            {
                "id": 5836309,
                "filename": "salmontools-result.tar.gz",
                "size_in_bytes": 191,
                "is_smashable": False,
                "is_qc": True,
                "sha1": "77ea7dd477fa9a9ec5f33bc91f46c0e224edeb92",
                "s3_bucket": "data-refinery-s3-circleci-prod",
                "s3_key": "gqo8hxhqm3tcz4c2x9sthhw5_salmontools-result.tar.gz",
                "created_at": "2020-10-01T20:23:41.661425Z",
                "last_modified": "2020-10-01T20:23:42.378884Z",
            }
        ],
        "organism_index": None,
        "time_start": "2020-10-01T20:23:06.885498Z",
        "time_end": "2020-10-01T20:23:41.030134Z",
        "created_at": "2020-10-01T20:23:41.655614Z",
        "last_modified": "2020-10-01T20:23:41.655614Z",
    },
}

computed_file_2 = {
    "id": 2,
    "filename": "salmontools-result.tar.gz",
    "samples": [
        {
            "accession_code": "SRR067392",
            "platform_name": "Illumina Genome Analyzer II",
            "pretty_platform": "Illumina Genome Analyzer II (IlluminaGenomeAnalyzerII)",
            "technology": "RNA-SEQ",
            "is_processed": True,
        }
    ],
    "size_in_bytes": 191,
    "is_qn_target": False,
    "is_smashable": False,
    "is_qc": True,
    "is_compendia": False,
    "quant_sf_only": False,
    "compendium_version": None,
    "sha1": "77ea7dd477fa9a9ec5f33bc91f46c0e224edeb92",
    "s3_bucket": "data-refinery-s3-circleci-prod",
    "s3_key": "gqo8hxhqm3tcz4c2x9sthhw5_salmontools-result.tar.gz",
    "s3_url": "https://s3.amazonaws.com/data-refinery-s3-circleci-prod/gqo8hxhqm3tcz4c2x9sthhw5_salmontools-result.tar.gz",
    "created_at": "2020-10-01T20:23:41.661425Z",
    "last_modified": "2020-10-01T20:23:42.378884Z",
    "result": {
        "id": 2784485,
        "commands": [
            "salmontools extract-unmapped -u /home/user/data_store/processor_job_29708301/SRR067392_output/aux_info/unmapped_names.txt -o /home/user/data_store/processor_job_29708301/salmontools/unmapped_by_salmon -r /home/user/data_store/processor_job_29708301/SRR067392.sra"
        ],
        "processor": {
            "id": 607,
            "name": "Salmontools",
            "version": "v1.39.5",
            "docker_image": "dr_salmon",
            "environment": {
                "os_pkg": {
                    "g++": "4:5.3.1-1ubuntu1",
                    "cmake": "3.5.1-1ubuntu3",
                    "python3": "3.5.1-3",
                    "python3-pip": "8.1.1-2ubuntu0.4",
                },
                "python": {"Django": "2.2.13", "data-refinery-common": "1.39.5"},
                "cmd_line": {"salmontools --version": "Salmon Tools 0.1.0"},
                "os_distribution": "Ubuntu 16.04.6 LTS",
            },
        },
        "is_ccdl": True,
        "annotations": [],
        "files": [
            {
                "id": 5836309,
                "filename": "salmontools-result.tar.gz",
                "size_in_bytes": 191,
                "is_smashable": False,
                "is_qc": True,
                "sha1": "77ea7dd477fa9a9ec5f33bc91f46c0e224edeb92",
                "s3_bucket": "data-refinery-s3-circleci-prod",
                "s3_key": "gqo8hxhqm3tcz4c2x9sthhw5_salmontools-result.tar.gz",
                "created_at": "2020-10-01T20:23:41.661425Z",
                "last_modified": "2020-10-01T20:23:42.378884Z",
            }
        ],
        "organism_index": None,
        "time_start": "2020-10-01T20:23:06.885498Z",
        "time_end": "2020-10-01T20:23:41.030134Z",
        "created_at": "2020-10-01T20:23:41.655614Z",
        "last_modified": "2020-10-01T20:23:41.655614Z",
    },
}

result = {
    "id": 2784485,
    "commands": [
        "salmontools extract-unmapped -u /home/user/data_store/processor_job_29708301/SRR067392_output/aux_info/unmapped_names.txt -o /home/user/data_store/processor_job_29708301/salmontools/unmapped_by_salmon -r /home/user/data_store/processor_job_29708301/SRR067392.sra"
    ],
    "processor": {
        "id": 607,
        "name": "Salmontools",
        "version": "v1.39.5",
        "docker_image": "dr_salmon",
        "environment": {
            "os_pkg": {
                "g++": "4:5.3.1-1ubuntu1",
                "cmake": "3.5.1-1ubuntu3",
                "python3": "3.5.1-3",
                "python3-pip": "8.1.1-2ubuntu0.4",
            },
            "python": {"Django": "2.2.13", "data-refinery-common": "1.39.5"},
            "cmd_line": {"salmontools --version": "Salmon Tools 0.1.0"},
            "os_distribution": "Ubuntu 16.04.6 LTS",
        },
    },
    "is_ccdl": True,
    "annotations": [],
    "files": [
        {
            "id": 5836309,
            "filename": "salmontools-result.tar.gz",
            "size_in_bytes": 191,
            "is_smashable": False,
            "is_qc": True,
            "sha1": "77ea7dd477fa9a9ec5f33bc91f46c0e224edeb92",
            "s3_bucket": "data-refinery-s3-circleci-prod",
            "s3_key": "gqo8hxhqm3tcz4c2x9sthhw5_salmontools-result.tar.gz",
            "created_at": "2020-10-01T20:23:41.661425Z",
            "last_modified": "2020-10-01T20:23:42.378884Z",
        }
    ],
    "organism_index": None,
    "time_start": "2020-10-01T20:23:06.885498Z",
    "time_end": "2020-10-01T20:23:41.030134Z",
    "created_at": "2020-10-01T20:23:41.655614Z",
    "last_modified": "2020-10-01T20:23:41.655614Z",
}

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [computed_file_1]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [computed_file_2]}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/computed_files/1/":
        return MockResponse(computed_file_1, url)

    if url == "https://api.refine.bio/v1/computed_files/2/":
        return MockResponse(computed_file_2, url)

    if url == "https://api.refine.bio/v1/computed_files/0/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/computed_files/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/computed_files/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)

    if url == "https://api.refine.bio/v1/computational_results/2784485/":
        return MockResponse(result, url)


class ComputedFileTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computed_file_get(self, mock_request):
        result = pyrefinebio.ComputedFile.get(1)
        self.assertObject(result, computed_file_1)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computed_file_500(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.ServerError):
            pyrefinebio.ComputedFile.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computed_file_get_404(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.NotFound):
            pyrefinebio.ComputedFile.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computed_file_search_no_filters(self, mock_request):
        results = pyrefinebio.ComputedFile.search()

        self.assertObject(results[0], computed_file_1)
        self.assertObject(results[1], computed_file_2)

    def test_computed_file_search_with_filters(self):
        filtered_results = pyrefinebio.ComputedFile.search(is_compendia=True, quant_sf_only=False)

        for result in filtered_results:
            self.assertTrue(result.is_compendia)
            self.assertFalse(result.quant_sf_only)

    def test_computed_file_search_with_invalid_filters(self):
        with self.assertRaises(pyrefinebio.exceptions.InvalidFilters):
            pyrefinebio.ComputedFile.search(foo="bar")

    # just mock download - it's already tested in depth in test_dataset
    @patch("pyrefinebio.computed_file.download_file")
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computed_file_download(self, mock_request, mock_download):
        result = pyrefinebio.ComputedFile.get(1)

        result.download("test-path")

        mock_download.assert_called_with("test_download_url", os.path.abspath("test-path"), True)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_computed_file_download_no_url(self, mock_request):
        result = pyrefinebio.ComputedFile.get(2)  # 2 has no download_url

        with self.assertRaises(pyrefinebio.exceptions.DownloadError) as de:
            result.download("test-path")

    @patch("pyrefinebio.computed_file.shutil.unpack_archive")
    def test_computed_file_extract(self, mock_unpack):
        cf = pyrefinebio.ComputedFile()

        cf._downloaded_path = "foo"

        cf.extract()

        mock_unpack.assert_called_with("foo")

    def test_computed_file_samples_are_fully_populated(self):
        # ComputedFile.Sample looks like this when retrieved from the API
        # We should make sure that the other sample properties are populated properly
        # {
        #     "accession_code": "SRR067392",
        #     "platform_name": "Illumina Genome Analyzer II",
        #     "pretty_platform": "Illumina Genome Analyzer II (IlluminaGenomeAnalyzerII)",
        #     "technology": "RNA-SEQ",
        #     "is_processed": True
        # }

        cfs = pyrefinebio.ComputedFile.search()
        cf = next(cf for cf in cfs if cf.samples)

        self.assertIsNotNone(cf.samples[0].experiment_accession_codes)
        self.assertIsNotNone(cf.samples[0].source_database)
        self.assertIsNotNone(cf.samples[0].organism)
        self.assertIsNotNone(cf.samples[0].platform_accession_code)
        self.assertIsNotNone(cf.samples[0].pretty_platform)
        self.assertIsNotNone(cf.samples[0].manufacturer)



from pprint import pformat


class Base(object):
    """Base class that all pyrefinebio classes inherit from.

    Contains helpful methods that relate to all classes.
    """

    def __init__(self, identifier=None):
        self._identifier = identifier
        self._fetched = False

    def __str__(self):
        return pformat(self._to_dict())

    def __eq__(self, other):
        try:
            return self._identifier == other._identifier
        except AttributeError:
            return False

    def _to_dict(self):
        expanded = {}

        for key, value in self.__dict__.items():
            if isinstance(value, Base):
                expanded[key] = value._to_dict()
            elif not key.startswith("_"):
                expanded[key] = value

        return expanded

    def __getattribute__(self, attr):
        if (
            not attr.startswith("_")
            and (
                object.__getattribute__(self, attr) is None
                or object.__getattribute__(self, attr) == []
            )
            and not self._fetched
        ):
            # if somehow the identifier isn't set we can't fetch
            # this will only happen with Datasets which can be created by non-api calls
            if self._identifier:
                result = self.get(self._identifier)
                for key, value in result.__dict__.items():
                    setattr(self, key, value)
                self._fetched = True

        return object.__getattribute__(self, attr)



from pyrefinebio import computational_result as prb_computational_result
from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.util import create_paginated_list, parse_date


class TranscriptomeIndex(Base):
    """Transcriptome Index.

    get a TranscriptomeIndex by id

        >>> import pyrefinebio
        >>> id = 1
        >>> t_index = pyrefinebio.TranscriptomeIndex.get(id)

    Retrieve a list of TranscriptomeIndex based on filters

        >>> import pyrefinebio
        >>> t_indecies = pyrefinebio.TranscriptomeIndex.search(organism_name="GORILLA")
    """

    def __init__(
        self,
        id=None,
        assembly_name=None,
        organism_name=None,
        database_name=None,
        release_version=None,
        index_type=None,
        salmon_version=None,
        download_url=None,
        result_id=None,
        last_modified=None,
    ):
        super().__init__(identifier=id)

        self.id = id
        self.assembly_name = assembly_name
        self.organism_name = organism_name
        self.database_name = database_name
        self.release_version = release_version
        self.index_type = index_type
        self.salmon_version = salmon_version
        self.download_url = download_url
        self.result_id = result_id
        self.result = prb_computational_result.ComputationalResult(id=result_id)
        self.last_modified = parse_date(last_modified)

    @classmethod
    def get(cls, id):
        """Retrieve a TranscriptomeIndex based on id

        Returns:
            TranscriptomeIndex

        Parameters:
            id (int): the id for the TranscriptomeIndex you want to get
        """
        response = get_by_endpoint("transcriptome_indices/" + str(id)).json()
        return TranscriptomeIndex(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of TranscriptomeIndex based on various filters

        Returns:
            list of TranscriptomeIndex

        Keyword Arguments:
            salmon_version (str): filter based on the TranscriptomeIndex's salmon version eg. salmon 0.13.1

            index_type (str): `TRANSCRIPTOME_LONG` or `TRANSCRIPTOME_SHORT`.
                              If the average read length of the RNA-Seq sample you want to process
                              is greater than 75 base pairs, use `TRANSCRIPTOME_LONG` otherwise use
                              `TRANSCRIPTOME_SHORT`

            ordering (str): which field to use when ordering the results.

            limit (int): number of results to return per page.

            offset (int): the initial index from which to return the results.

            organism__name (str): filter based on the name of the Organism associated
                                  with the TranscriptomeIndex

            length (str): short hand for index_type eg. `short` or `long`
                          see `index_type` for more information
        """
        response = get_by_endpoint("transcriptome_indices", params=kwargs)
        return create_paginated_list(cls, response)



import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

processor_1 = {
    "id": 1,
    "name": "test processor 1",
    "version": "v1.0.1",
    "docker_image": "test_image_1",
    "environment": {
        "os_pkg": {
            "g++": "4:5.3.1-1ubuntu1",
            "python3": "3.5.1-3",
            "python3-pip": "8.1.1-2ubuntu0.4",
        },
        "python": {"Django": "2.0.6", "data-refinery-common": "1.0.1"},
        "cmd_line": {
            "salmon --version": "salmon 0.9.1",
            "rsem-calculate-expression --version": "Current version: RSEM v1.3.1",
        },
        "os_distribution": "Ubuntu 16.04.5 LTS",
    },
}

processor_2 = {
    "id": 2,
    "name": "test processor 2",
    "version": "v1.0.1",
    "docker_image": "test_image_2",
    "environment": {
        "os_pkg": {"python3": "3.5.1-3", "python3-pip": "8.1.1-2ubuntu0.4"},
        "python": {"Django": "2.0.6", "data-refinery-common": "1.0.1"},
        "cmd_line": {"fasterq-dump --version": "fasterq-dump : 2.9.1"},
        "os_distribution": "Ubuntu 16.04.5 LTS",
    },
}

search_1 = {"count": 2, "next": "search_2", "previous": None, "results": [processor_1]}

search_2 = {"count": 2, "next": None, "previous": "search_1", "results": [processor_2]}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/processors/1/":
        return MockResponse(processor_1, url)

    if url == "https://api.refine.bio/v1/processors/0/":
        return MockResponse(None, url, status=404)

    if url == "https://api.refine.bio/v1/processors/500/":
        return MockResponse(None, url, status=500)

    if url == "https://api.refine.bio/v1/processors/":
        return MockResponse(search_1, "search_2")

    if url == "search_2":
        return MockResponse(search_2, url)


class ProcessorTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_processor_get(self, mock_request):
        result = pyrefinebio.Processor.get(1)
        self.assertObject(result, processor_1)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_processor_500(self, mock_request):
        with self.assertRaises(Exception):
            pyrefinebio.Processor.get(500)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_processor_get_404(self, mock_request):
        with self.assertRaises(Exception):
            pyrefinebio.Processor.get(0)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_processor_search_no_filters(self, mock_request):
        results = pyrefinebio.Processor.search()

        self.assertObject(results[0], processor_1)
        self.assertObject(results[1], processor_2)



import unittest
import pyrefinebio
import os
from pyrefinebio.config import Config
from unittest.mock import Mock, patch

from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse


class HighLevelFunctionTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.dataset.download_file")
    @patch("pyrefinebio.dataset.get_by_endpoint")
    @patch("pyrefinebio.dataset.post_by_endpoint")
    def test_download_dataset_dataset_dict(self, mock_post, mock_get, mock_download):
        mr = MockResponse(
            {"id": "test_dataset", "is_processed": True, "download_url": "test_url"}, "url"
        )

        mock_get.return_value = mr
        mock_post.return_value = mr

        dataset_dict = {"test": ["foo", "bar"]}

        pyrefinebio.download_dataset("test", "foo@test.com", dataset_dict=dataset_dict)

        mock_post.assert_called_with(
            "dataset",
            payload={
                "data": dataset_dict,
                "aggregate_by": "EXPERIMENT",
                "email_address": "foo@test.com",
                "start": True,
                "quantile_normalize": True,
                "quant_sf_only": False,
                "notify_me": False,
            },
        )

        mock_download.assert_called_with("test_url", os.path.abspath("test"), True)

    @patch("pyrefinebio.dataset.download_file")
    @patch("pyrefinebio.dataset.get_by_endpoint")
    @patch("pyrefinebio.dataset.post_by_endpoint")
    def test_download_dataset_notify_me(self, mock_post, mock_get, mock_download):
        mr = MockResponse(
            {"id": "test_dataset", "is_processed": True, "download_url": "test_url"}, "url"
        )

        mock_get.return_value = mr
        mock_post.return_value = mr

        dataset_dict = {"test": ["foo", "bar"]}

        pyrefinebio.download_dataset(
            "test", "foo@test.com", dataset_dict=dataset_dict, notify_me=True
        )

        mock_post.assert_called_with(
            "dataset",
            payload={
                "data": dataset_dict,
                "aggregate_by": "EXPERIMENT",
                "email_address": "foo@test.com",
                "start": True,
                "quantile_normalize": True,
                "quant_sf_only": False,
                "notify_me": True,
            },
        )

        mock_download.assert_called_with("test_url", os.path.abspath("test"), True)

    @patch("pyrefinebio.dataset.download_file")
    @patch("pyrefinebio.dataset.get_by_endpoint")
    @patch("pyrefinebio.dataset.post_by_endpoint")
    def test_download_dataset_experiments(self, mock_post, mock_get, mock_download):
        mr = MockResponse(
            {"id": "test_dataset", "is_processed": True, "download_url": "test_url"}, "url"
        )

        mock_get.return_value = mr
        mock_post.return_value = mr

        experiment2 = pyrefinebio.Experiment(accession_code="test-experiment-2")

        pyrefinebio.download_dataset(
            "test", "foo@test.com", experiments=["test-experiment-1", experiment2]
        )

        mock_post.assert_called_with(
            "dataset",
            payload={
                "data": {"test-experiment-1": ["ALL"], "test-experiment-2": ["ALL"]},
                "aggregate_by": "EXPERIMENT",
                "email_address": "foo@test.com",
                "start": True,
                "quantile_normalize": True,
                "quant_sf_only": False,
                "notify_me": False,
            },
        )

        mock_download.assert_called_with("test_url", os.path.abspath("test"), True)

    @patch("pyrefinebio.dataset.shutil.unpack_archive")
    @patch("pyrefinebio.dataset.download_file")
    @patch("pyrefinebio.dataset.get_by_endpoint")
    @patch("pyrefinebio.dataset.post_by_endpoint")
    def test_download_dataset_extract(self, mock_post, mock_get, mock_download, mock_unpack):
        mr = MockResponse(
            {"id": "test_dataset", "is_processed": True, "download_url": "test_url"}, "url"
        )

        mock_get.return_value = mr
        mock_post.return_value = mr

        dataset_dict = {"test": ["foo", "bar"]}

        pyrefinebio.download_dataset("./", "foo@test.com", dataset_dict=dataset_dict, extract=True)

        expected_path = os.path.abspath("dataset-test_dataset.zip")

        mock_download.assert_called_with("test_url", expected_path, True)
        mock_unpack.assert_called_with(expected_path)

    def test_download_dataset_both(self):
        with self.assertRaises(pyrefinebio.exceptions.DownloadError):
            pyrefinebio.download_dataset(
                "test", "foo@test.com", dataset_dict={"bad"}, experiments=["bad"]
            )

    @patch("pyrefinebio.compendia.download_file")
    @patch("pyrefinebio.compendia.get_by_endpoint")
    def test_download_compendium(self, mock_get, mock_download):
        mock_get.return_value = MockResponse(
            {
                "count": 1,
                "next": None,
                "previous": None,
                "results": [{"computed_file": {"download_url": "test_url"}}],
            },
            "url",
        )

        pyrefinebio.download_compendium("test", "test_organism")

        mock_get.assert_called_with(
            "compendia",
            params={
                "primary_organism__name": "test_organism",
                "quant_sf_only": False,
                "latest_version": True,
            },
        )

        mock_download.assert_called_with("test_url", os.path.abspath("test"), True)

    @patch("pyrefinebio.compendia.shutil.unpack_archive")
    @patch("pyrefinebio.compendia.download_file")
    @patch("pyrefinebio.compendia.get_by_endpoint")
    def test_download_compendium_extract(self, mock_get, mock_download, mock_unpack):
        mock_get.return_value = MockResponse(
            {
                "count": 1,
                "next": None,
                "previous": None,
                "results": [{"id": 123, "computed_file": {"download_url": "test_url"}}],
            },
            "url",
        )

        pyrefinebio.download_compendium("./", "test_organism", extract=True)

        expected_path = os.path.abspath("compendium-123.zip")

        mock_download.assert_called_with("test_url", expected_path, True)
        mock_unpack.assert_called_with(expected_path)

    def test_download_compendium_no_results(self):
        with self.assertRaises(pyrefinebio.exceptions.DownloadError):
            pyrefinebio.download_compendium("test", "this-will-have-no-results")

    @patch("pyrefinebio.computed_file.get_by_endpoint")
    @patch("pyrefinebio.compendia.get_by_endpoint")
    def test_download_compendium_no_token(self, mock_compendia_get, mock_computed_file_get):
        mock_compendia_get.return_value = MockResponse(
            {
                "count": 1,
                "next": None,
                "previous": None,
                "results": [{"computed_file": {"download_url": None}}],
            },
            "url",
        )

        mock_computed_file_get.return_value = MockResponse({"download_url": None}, "url")

        with self.assertRaises(pyrefinebio.exceptions.DownloadError):
            pyrefinebio.download_compendium("test", "test_organism")

    @patch("pyrefinebio.high_level_functions.Token")
    @patch("pyrefinebio.high_level_functions.input")
    def test_create_token_with_prompt(self, mock_input, mock_token):
        mock_input.return_value = "y"

        mock_created_token = pyrefinebio.create_token()

        self.assertTrue(mock_token.called)

        self.assertTrue(mock_created_token.agree_to_terms_and_conditions.called)
        self.assertTrue(mock_created_token.save_token.called)

    @patch("pyrefinebio.high_level_functions.Token")
    @patch("pyrefinebio.high_level_functions.input")
    def test_create_token_activate_no_save(self, mock_input, mock_token):
        mock_created_token = pyrefinebio.create_token(agree_to_terms=True, save_token=False)

        self.assertTrue(mock_token.called)

        self.assertTrue(mock_created_token.agree_to_terms_and_conditions.called)
        self.assertFalse(mock_created_token.save_token.called)



import shutil

from pyrefinebio import computational_result as prb_computational_result, sample as prb_sample
from pyrefinebio.api_interface import download_file, get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.exceptions import DownloadError, MissingFile
from pyrefinebio.util import create_paginated_list, expand_path, parse_date


class ComputedFile(Base):
    """Computed File.

    ComputedFiles are representation of files created by data-refinery processes.

    Retrieve a ComputedFile based on id

        >>> import pyrefinebio
        >>> id = 1
        >>> file = pyrefinebio.ComputedFile.get(id)

    Retrieve a list of ComputedFile based on filters

        >>> import pyrefinebio
        >>> files = pyrefinebio.ComputedFile.search(is_compendia=True, is_public=True)
    """

    def __init__(
        self,
        id=None,
        filename=None,
        samples=None,
        size_in_bytes=None,
        is_qn_target=None,
        is_smashable=None,
        is_qc=None,
        is_compendia=None,
        quant_sf_only=None,
        compendium_version=None,
        compendia_organism_name=None,
        sha1=None,
        s3_bucket=None,
        s3_key=None,
        s3_url=None,
        download_url=None,
        created_at=None,
        last_modified=None,
        result=None,
    ):
        super().__init__(identifier=id)

        self.id = id
        self.filename = filename
        self.samples = [prb_sample.Sample(**sample) for sample in samples] if samples else []
        self.size_in_bytes = size_in_bytes
        self.is_qn_target = is_qn_target
        self.is_smashable = is_smashable
        self.is_qc = is_qc
        self.is_compendia = is_compendia
        self.quant_sf_only = quant_sf_only
        self.compendium_version = compendium_version
        self.compendia_organism_name = compendia_organism_name
        self.sha1 = sha1
        self.s3_bucket = s3_bucket
        self.s3_key = s3_key
        self.s3_url = s3_url
        self.download_url = download_url
        self.created_at = parse_date(created_at)
        self.last_modified = parse_date(last_modified)
        self.result = prb_computational_result.ComputationalResult(**(result)) if result else None

        self._downloaded_path = None

    @classmethod
    def get(cls, id):
        """Retrieve a specific ComputedFile based on id

        Returns:
            ComputedFile

        Parameters:
            id (int): the id for the ComputedFile you want to get
        """
        response = get_by_endpoint("computed_files/" + str(id)).json()
        return ComputedFile(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of a ComputedFiles based on filters

        Returns:
            list of ComputedFile

        Keyword Arguments:

            id (int): filter based on the id of the ComputedFile

            samples (str): filter based on the accession code for Samples related to the ComputedFile

            is_qn_target (bool): filter based on if the ComputedFile is a qn target

            is_smashable (bool): filter based on if the ComputedFile can be added to a normalized
                                 Dataset

            is_qc (bool): filter based on if the ComputedFile contains data about
                          the quality control of a result rather than data about the
                          result itself

            is_compendia (bool): filter based on if the ComputedFile is part of a compendium

            quant_sf_only (bool): filter based on if the Samples associated with the ComputedFile
                                  are RNA-seq only

            svd_algorithm (str): filter based on the SVD algorithm used for the ComputedFile

            compendium_version (int): filter based on the compendium version of the ComputedFile

            created_at (str): filter based on the time that the ComputedFile was created

            last_modified (str): filter based on the time that the ComputedFile was last modified

            result__id (int): filter based on the id of the ComputationalResult associated with the
                              ComputedFile

            ordering (str): which field to use when ordering the results

            limit (int): number of results to return per page

            offset (int): the initial index from which to return the results
        """
        response = get_by_endpoint("computed_files", params=kwargs)
        return create_paginated_list(cls, response)

    def download(self, path, prompt=True):
        """Download a ComputedFile

        Returns:
            ComputedFile

        Parameters:
            path (str): the path that the ComputedFile should be downloaded to

            prompt (bool): if true, will prompt before downloading files bigger than 1GB
        """
        if not self.download_url:
            raise DownloadError(
                "ComputedFile",
                "Download url not found - make sure you have set up and activated your Token. "
                "You can create and activate a new token using pyrefinebio.create_token(). "
                "See documentation for advanced usage: https://alexslemonade.github.io/refinebio-py/token.html",
            )

        full_path = expand_path(path, "computedfile-" + str(self.id) + ".zip")

        download_file(self.download_url, full_path, prompt)

        self._downloaded_path = full_path

        return self

    def extract(self):
        """Extract a downloaded ComputedFile

        Returns:
            ComputedFile
        """
        if not self._downloaded_path:
            raise MissingFile(
                "ComputedFile downloaded file",
                "Make sure you have successfully downloaded the ComputedFile before extracting.",
            )

        shutil.unpack_archive(self._downloaded_path)
        return self



import shutil

from pyrefinebio import experiment as prb_experiment, sample as prb_sample
from pyrefinebio.api_interface import (
    download_file,
    get_by_endpoint,
    post_by_endpoint,
    put_by_endpoint,
)
from pyrefinebio.base import Base
from pyrefinebio.exceptions import DownloadError, MissingFile
from pyrefinebio.util import expand_path, parse_date


class Dataset(Base):
    """Dataset

    Datasets are collections of experiments and their samples.
    A Dataset needs to be constructed and then processed before it can be downloaded.
    Downloading a Dataset requires an activated API token. See `pyrefinebio.Token` for more details

    Create and save a Dataset

        >>> import pyrefinebio
        >>> dataset = pyrefinebio.Dataset(email_address="example@refine.bio", data={"SRP003819": ["SRR069230"]})
        >>> dataset = dataset.save()

    Get a Dataset that has been saved

        >>> import pyrefinebio
        >>> id = "dataset id <guid>"
        >>> dataset = pyrefinebio.Dataset.get(id)

    Start processing a Dataset

        >>> import pyrefinebio
        >>> dataset = pyrefinebio.Dataset(...)
        >>> dataset.process()

    Check if a Dataset is finished processing

        >>> import pyrefinebio
        >>> dataset = pyrefinebio.Dataset(...)
        >>> dataset.process()
        >>> dataset.check()

    Download a processed Dataset

        >>> import pyrefinebio
        >>> dataset = pyrefinebio.Dataset(...)
        >>> dataset.process()
        >>> dataset.download("~/datasets/my_dataset.zip")
    """

    def __init__(
        self,
        id=None,
        data=None,
        aggregate_by=None,
        scale_by=None,
        is_processing=None,
        is_processed=None,
        is_available=None,
        has_email=None,
        email_address=None,
        email_ccdl_ok=None,
        expires_on=None,
        s3_bucket=None,
        s3_key=None,
        success=None,
        failure_reason=None,
        created_at=None,
        last_modified=None,
        start=None,
        size_in_bytes=None,
        sha1=None,
        quantile_normalize=None,
        quant_sf_only=None,
        svd_algorithm=None,
        download_url=None,
        # Emails should be opt-in https://github.com/AlexsLemonade/refinebio-py/issues/58
        notify_me=False,
    ):
        super().__init__(identifier=id)

        self.id = id
        self.data = data
        self.aggregate_by = aggregate_by
        self.scale_by = scale_by
        self.is_processing = is_processing
        self.is_processed = is_processed
        self.is_available = is_available
        self.has_email = has_email
        self.email_address = email_address
        self.email_ccdl_ok = email_ccdl_ok
        self.expires_on = parse_date(expires_on)
        self.s3_bucket = s3_bucket
        self.s3_key = s3_key
        self.success = success
        self.failure_reason = failure_reason
        self.created_at = parse_date(created_at)
        self.last_modified = parse_date(last_modified)
        self.start = start
        self.size_in_bytes = size_in_bytes
        self.sha1 = sha1
        self.quantile_normalize = quantile_normalize
        self.quant_sf_only = quant_sf_only
        self.svd_algorithm = svd_algorithm
        self.download_url = download_url
        self.notify_me = notify_me

        self._downloaded_path = None

    @classmethod
    def get(cls, id):
        """Retrieve a specific Dataset based on id

        Returns:
            Dataset

        Parameters:
            id (str): the guid id for the computed file you want to get
        """
        response = get_by_endpoint("dataset/" + id).json()
        return Dataset(**response)

    def add_samples(self, experiment, samples=["ALL"]):
        """Add samples to a dataset

        Returns:
            Dataset

        Parameters:
            experiment (str): accession code for the Experiment related to the Samples you
                              are adding to the dataset
                       (Experiment): Experiment object related to the Samples you are adding
                              the dataset

            samples (list): list of Sample objects or Sample accession codes for the samples
                            you are adding to the dataset
        """
        if self.is_processing or self.is_processed:
            print("Cannot add samples to a dataset that has been processed!")
            return

        # get accession codes if Experiment or Sample objects are passed in
        if isinstance(experiment, prb_experiment.Experiment):
            experiment = experiment.accession_code

        sample_accession_codes = []
        for sample in samples:
            if isinstance(sample, prb_sample.Sample):
                sample_accession_codes.append(sample.accession_code)
            else:
                sample_accession_codes.append(sample)

        samples = sample_accession_codes

        if self.data:
            self.data[experiment] = samples
        else:
            self.data = {experiment: samples}

        return self

    def save(self):
        """Save a Dataset

        In order for a dataset to be saved its `data` attribute must be properly set.
        The `data` attribute should be a dict with experiment accession codes as the keys
        and lists of sample accession codes as the values. If you want all samples associated
        with the experiment, you can use the value `"ALL"`.

        Example:
            >>> data = {
            >>>     "SRP003819": [
            >>>         "SRR069230",
            >>>         "SRR069231"
            >>>     ],
            >>>     "SRP003820": ["ALL"]
            >>> }

        Returns:
            Dataset
        """
        body = {}
        body["data"] = self.data
        if self.aggregate_by is not None:
            body["aggregate_by"] = self.aggregate_by
        if self.scale_by is not None:
            body["scale_by"] = self.scale_by
        if self.email_address is not None:
            body["email_address"] = self.email_address
        if self.email_ccdl_ok is not None:
            body["email_ccdl_ok"] = self.email_ccdl_ok
        if self.start is not None:
            body["start"] = self.start
        if self.quantile_normalize is not None:
            body["quantile_normalize"] = self.quantile_normalize
        if self.quant_sf_only is not None:
            body["quant_sf_only"] = self.quant_sf_only
        if self.svd_algorithm is not None:
            body["svd_algorithm"] = self.svd_algorithm
        if self.notify_me is not None:
            body["notify_me"] = self.notify_me

        if self.id is not None:
            response = put_by_endpoint("dataset/" + self.id, payload=body).json()
        else:
            response = post_by_endpoint("dataset", payload=body).json()

        # add fields that aren't returned by the api
        response["email_address"] = self.email_address
        response["email_ccdl_ok"] = self.email_ccdl_ok

        for key, value in response.items():
            setattr(self, key, value)

        return self

    def process(self):
        """Start processing a Dataset

        In order for a Dataset to be processed, its `data` and `email_address` attributes
        must be properly set.

        Returns:
            void
        """
        self.start = True
        response = self.save()
        self.is_processing = response.is_processing

    def check(self):
        """Check to see if a Dataset has finished processing

        Returns:
            bool
        """

        response = self.get(self.id)
        self.is_processing = response.is_processing
        self.is_processed = response.is_processed
        return response.is_processed

    def download(self, path, prompt=True):
        """Download a processed Dataset

        The path that the dataset is downloaded to is stored in `_downloaded_path`

        Returns:
            Dataset

        Parameters:
            path (str): the path that the Dataset should be downloaded to

            prompt (bool): if true, will prompt before downloading files bigger than 1GB
        """
        download_url = self.download_url or self.get(self.id).download_url

        if not download_url:
            if self.check():
                raise DownloadError(
                    "Dataset",
                    "Download url not found - make sure you have set up and activated your Token. "
                    "You can create and activate a new token using pyrefinebio.create_token(). "
                    "See documentation for advanced usage: https://alexslemonade.github.io/refinebio-py/token.html",
                )
            else:
                raise DownloadError(
                    "Dataset",
                    "Download url not found - you must process the Dataset before downloading.",
                )

        full_path = expand_path(path, "dataset-" + str(self.id) + ".zip")

        download_file(download_url, full_path, prompt)

        self._downloaded_path = full_path

        return self

    def extract(self):
        """Extract a downloaded Dataset

        Returns:
            Dataset
        """
        if not self._downloaded_path:
            raise MissingFile(
                "Dataset downloaded file"
                "Make sure you have successfully downloaded the Dataset before extracting."
            )

        shutil.unpack_archive(self._downloaded_path)
        return self



import re
import time
from datetime import datetime, timedelta

import pyrefinebio
from pyrefinebio import Compendium, Dataset, Token
from pyrefinebio.exceptions import DownloadError


def help(entity=None):
    """Help

    Prints out information about pyrefinebio's classes and functions.
    To get information about a class method, pass in the class name and method name
    separated by either a space or a `.`

    Examples:
        getting information about classes:

        >>> pyrefinebio.help("Sample")

        getting information about class methods:

        >>> pyrefinebio.help("Sample.get")
        or
        >>> pyrefinebio.help("Sample get")

        getting information about functions:

        >>> pyrefinebio.help("download_dataset")
    """

    if not entity:
        print(help.__doc__)
        return

    try:
        final_entity = pyrefinebio
        for attr in re.split("[. ]", entity):
            final_entity = getattr(final_entity, attr)
        print(final_entity.__doc__)
    except:
        print("could not find class or attribute: ", entity)


def download_dataset(
    path,
    email_address,
    dataset_dict=None,
    experiments=None,
    aggregation="EXPERIMENT",
    transformation="NONE",
    skip_quantile_normalization=False,
    quant_sf_only=False,
    timeout=None,
    extract=False,
    prompt=True,
    notify_me=False,
):
    """download_dataset

    Automatically constructs a Dataset, processes it, waits for it
    to finish processing, then downloads it to the path specified.

    Returns:
        Dataset

    Parameters:
        path (str): the path that the Dataset should be downloaded to

        email_address (str): the email address that will be contacted with info
                                related to the dataset

        dataset_dict (dict): a fully formed Dataset `data` attribute in the form:
                                {
                                    "Experiment": [
                                        "Sample",
                                        "Sample"
                                    ]
                                }
                                use this parameter if you want to specify specific Samples
                                for your dataset
                                each part of the dict can be a pyrefinebio object or an accession
                                code as a string

        experiments (list): a list of Experiments that should be added to the dataset
                            use this parameter if you only care about the Experiments - all
                            available samples related to each Experiment will be added
                            the list can contain Experiment objects or accession codes as strings

        aggregation (str): how the Dataset should be aggregated - by `EXPERIMENT`, by `SPECIES`, or by `ALL`

        transformation (str): the transformation for the dataset - `NONE`, `MINMAX`, or `STANDARD`

        skip_quantile_normalization (bool): control whether or not the dataset should skip quantile
                                            normalization for RNA-seq Samples

        quant_sf_only (bool): include only quant.sf files in the generated dataset.

        timeout (datetime.timedelta): if specified the function will return None after timeout is reached.

        extract (bool): if true, the downloaded zip file will be automatically extracted

        prompt (bool): if true, will prompt before downloading files bigger than 1GB

        notify_me (bool): if true, refine.bio will send you an email when the dataset has finished processing.
                          Defaults to False.
    """
    if dataset_dict and experiments:
        raise DownloadError(
            "Dataset",
            extra_info="You should either provide dataset_dict or experiments but not both",
        )

    print("Creating Dataset...")

    dataset = Dataset(
        email_address=email_address,
        aggregate_by=aggregation,
        quantile_normalize=(not skip_quantile_normalization),
        quant_sf_only=quant_sf_only,
        notify_me=notify_me,
    )

    if dataset_dict:
        dataset.data = dataset_dict

    if experiments:
        for experiment in experiments:
            dataset.add_samples(experiment)

    print("Processing Dataset...")

    dataset.process()

    start_time = datetime.now()
    while not dataset.check():
        if not dataset.is_processing and not dataset.is_processed:
            print(
                "Dataset failed to process. The system may be experiencing issues."
                " Please try again later."
            )
            return None
        elif timeout and datetime.now() - start_time > timeout:
            print(
                (
                    "Dataset not processed after {0}. The system may be experiencing issues"
                    " or your dataset may just be taking a while to process. You can check on its"
                    " progress at https://refine.bio/dataset/{1}"
                ).format(timeout, dataset.id)
            )
            return None

        time.sleep(5)

    print("Downloading Dataset...")

    dataset = dataset.download(path, prompt)

    if extract:
        print("Extracting Dataset...")
        dataset.extract()

    return dataset


def download_compendium(
    path, organism, version=None, quant_sf_only=False, extract=False, prompt=True
):
    """download_compendium

    Download a Compendium for the specified organism.

    Returns:
        Compendium

    Parameters:
        path (str): the path that the Compendium should be downloaded to

        organism (str): the name of the Organism for the Compendium you want to
                        download

        version (int): the version for the Compendium you want to download - None
                       for latest version

        quant_sf_only (bool): true for RNA-seq Sample Compendium results or False
                              for quantile normalized

        extract (bool): if true, the downloaded zip file will be automatically extracted

        prompt (bool): if true, will prompt before downloading files bigger than 1GB
    """
    search_params = {"primary_organism__name": organism, "quant_sf_only": quant_sf_only}

    if version:
        search_params["compendium_version"] = version
    else:
        search_params["latest_version"] = True

    print("Searching for Compendium...")

    compendium = Compendium.search(**search_params)

    if not compendium:
        raise DownloadError(
            "Compendium",
            extra_info="Could not find any Compendium with organism name, {0}, version {1}, and quant_sf_only, {2}".format(
                organism, version, quant_sf_only
            ),
        )

    print("Downloading Compendium...")

    compendium = compendium[0].download(path, prompt)

    if extract:
        print("Extracting Compendium...")
        compendium.extract()

    return compendium


def download_quantfile_compendium(path, organism, version=None, extract=False, prompt=True):
    """download_quantfile_compendium

    Download a Compendium for the specified organism.
    This function will always download RNA-seq Sample Compedium results.

    Returns:
        Compendium

    Parameters:
        path (str): the path that the Compendium should be downloaded to

        organism (str): the name of the Organism for the Compendium you want to
                        download

        version (int): the version for the Compendium you want to download - None
                       for latest version

        extract (bool): if true, the downloaded zip file will be automatically extracted

        prompt (bool): if true, will prompt before downloading files bigger than 1GB
    """
    return download_compendium(
        path, organism, version=version, quant_sf_only=True, extract=extract, prompt=prompt
    )


def create_token(agree_to_terms=None, save_token=None):
    """create_token

    Automatically creates a Token, activates it, and stores it to the Config file.

    By default, will prompt the user before activating and storing the created Token.

    Returns:
        Token

    Parameters:
        agree_to_terms (bool): if true, the token will be automatically activated without prompting.
                               If false, the token will not be activated. Leave as None to be prompted
                               before activating.

        save_token (bool): if true, the token will be automatically saved. If false, the token will
                           not be saved. Leave as None to be prompted before saving.
    """
    if agree_to_terms is None:
        print(
            "Please review the refine.bio Terms of Use: https://www.refine.bio/terms and Privacy Policy: https://www.refine.bio/privacy"
        )
        yn = input("Do you understand and accept both documents? (y/N)")
        agree_to_terms = yn.lower() in ("y", "yes")

    token = Token()

    if agree_to_terms:
        token.agree_to_terms_and_conditions()

    if save_token is None:
        yn = input("Would you like to save your Token to the Config file for future use? (y/N)")
        save_token = yn.lower() in ("y", "yes")

    if save_token:
        token.save_token()

    return token



from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base


class Platform(Base):
    """Platform.

    Retrieve a list of Platforms based on filters

        >>> import pyrefinebio
        >>> og_files = pyrefinebio.Platform.search()
    """

    def __init__(self, platform_accession_code=None, platform_name=None):
        self.platform_accession_code = platform_accession_code
        self.platform_name = platform_name

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of Platforms

        Returns:
            list of Platform

        Since there are no filters, this method always returns all Platforms
        """
        response = get_by_endpoint("platforms", params=kwargs).json()
        return [Platform(**platform) for platform in response]



from pyrefinebio.base import Base

from pyrefinebio.util import parse_date


class Annotation(Base):
    def __init__(
        self,
        id=None,
        data=None,
        is_ccdl=None,
        created_at=None,
        last_modified=None
    ):
        self.id = id
        self.data = data
        self.is_ccdl = is_ccdl
        self.created_at = parse_date(created_at)
        self.last_modified = parse_date(last_modified)



from requests.exceptions import HTTPError

class MockResponse:
    def __init__(self, json_data, url, status=200, headers=None):
        self.json_data = json_data
        self.url = url
        self.status_code = status
        self.headers = headers

    def json(self):
        return self.json_data
    
    def raise_for_status(self):
        if self.status_code != 200:
            raise HTTPError



from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base


class Institution(Base):
    """Institution

    Retrieve a list of Institutions

        >>> import pyrefinebio
        >>> institutions = pyrefinebio.Institution.search()
    """

    def __init__(self, submitter_institution=None):
        self.submitter_institution = submitter_institution

    @classmethod
    def search(cls):
        """Retrieve a list of Institutions

        Returns:
            list of Institution

        Since there are no filters, this method always gets all Institutions.
        """
        response = get_by_endpoint("institutions").json()
        return [Institution(**institution) for institution in response]



from pyrefinebio import (
    annotation as prb_annotation,
    computed_file as prb_computed_file,
    processor as prb_processor,
    transcriptome_index as prb_transcriptome_index,
)
from pyrefinebio.api_interface import get_by_endpoint
from pyrefinebio.base import Base
from pyrefinebio.util import create_paginated_list, parse_date


class ComputationalResult(Base):
    """Computational Result

    Retrieve a ComputationalResult based on id

        >>> import pyrefinebio
        >>> id = 1
        >>> result = pyrefinebio.ComputationalResult.get(id)

    Retrieve a list of ComputationalResult based on filters

        >>> import pyrefinebio
        >>> results = pyrefinebio.ComputationalResult.search(processor_id=4)
    """

    def __init__(
        self,
        id=None,
        commands=None,
        processor=None,
        is_ccdl=None,
        annotations=None,
        files=None,
        organism_index=None,
        time_start=None,
        time_end=None,
        created_at=None,
        last_modified=None,
    ):
        super().__init__(identifier=id)

        self.id = id
        self.commands = commands
        self.processor = prb_processor.Processor(**(processor)) if processor else None
        self.is_ccdl = is_ccdl
        self.annotations = (
            [prb_annotation.Annotation(**annotation) for annotation in annotations]
            if annotations
            else []
        )
        self.files = [prb_computed_file.ComputedFile(**file) for file in files] if files else []
        if organism_index:
            self.organism_index = prb_transcriptome_index.TranscriptomeIndex(**(organism_index))
        else:
            self.organism_index = None

        self.time_start = parse_date(time_start)
        self.time_end = parse_date(time_end)
        self.created_at = parse_date(created_at)
        self.last_modified = parse_date(last_modified)

    @classmethod
    def get(cls, id):
        """Retrieve a computational result based on its id.

        Returns:
            ComputationalResult

        Parameters:
            id (int): The id for the computational result to be retrieved.
        """
        response = get_by_endpoint("computational_results/" + str(id)).json()
        return ComputationalResult(**response)

    @classmethod
    def search(cls, **kwargs):
        """Retrieve a list of computational results based on filters.

        Returns:
            list of ComputationalResult

        Keyword Arguments:
            processor__id (int): id of the Processor that processed the result

            limit (int): number of results to return per page

            offset (int): the initial index from which to return the results
        """
        response = get_by_endpoint("computational_results", params=kwargs)
        return create_paginated_list(cls, response)



import os
from pathlib import Path

import yaml


class Config:
    """Config for pyrefinebio.

    Config values can be set by environment variables or in the Config file.

    The Config file's default location is `~/.refinebio.yaml`, but this location
    can be set by using the environment variable `REFINEBIO_CONFIG_FILE`

    pyrefinebio's configurable values are:

        token:
            The refine.bio API token that is used when making requests

            environment variable: `REFINEBIO_TOKEN`

        base_url:
            The base url for the refine.bio API that should be used
            when making requests. The default is `https://api.refine.bio/v1/`

            environment variable: `REFINEBIO_BASE_URL`

        api_max_calls_per_second:
            The refine.bio API calls per second limit. This defaults to the API's rate limit
            which is enforced per IP address.

            environment variable: `REFINEBIO_API_MAX_CALLS_PER_SECOND`

    These config values can be modified directly in code, but it recommended that you
    set them by using environment variables, by modifying them in Config file, or by
    using other class methods provided - like `pyrefinebio.Token` for example.

    example Config file:

    .. code-block:: shell

        token: foo-bar-baz
        base_url: https://api.refine.bio/v1/
        api_max_calls_per_second: 10
    """

    _instance = None

    def __new__(cls):
        """Create an instance of config

        Config values are loaded from environment variables first, then the config file,
        then they are set to defaults.

        returns:
            Config
        """
        if not cls._instance:
            cls._instance = super(Config, cls).__new__(cls)

            cls.config_file = (
                os.getenv("REFINEBIO_CONFIG_FILE") or str(Path.home()) + "/.refinebio.yaml"
            )

            config = {}

            if os.path.exists(cls.config_file):
                with open(cls.config_file) as config_file:
                    config = yaml.safe_load(config_file) or {}

            cls.token = os.getenv("REFINEBIO_TOKEN") or config.get("token", "")
            cls.base_url = os.getenv("REFINEBIO_BASE_URL") or config.get(
                "base_url", "https://api.refine.bio/v1/"
            )
            cls.api_max_calls_per_second = int(
                os.getenv("REFINEBIO_API_MAX_CALLS_PER_SECOND")
                or config.get("api_max_calls_per_second", 10)
            )

        return cls._instance

    def save(self):
        """Save the config to a file

        The default path for this file is `~/.refinebio.yaml`
        but it can be set using the environment variable `REFINEBIO_CONFIG_FILE`.
        """
        config = {
            "token": self.token,
            "base_url": self.base_url,
            "api_max_calls_per_second": self.api_max_calls_per_second,
        }

        with open(self.config_file, "w") as config_file:
            yaml.dump(config, config_file)



import unittest
from unittest.mock import Mock, patch

import pyrefinebio
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

item0 = {"id": 1, "name": "test-1", "version": 1, "docker_image": "test-1", "environment": "test-1"}

item1 = {"id": 2, "name": "test-2", "version": 1, "docker_image": "test-2", "environment": "test-2"}

item2 = {"id": 3, "name": "test-3", "version": 1, "docker_image": "test-3", "environment": "test-3"}

item3 = {"id": 4, "name": "test-4", "version": 1, "docker_image": "test-4", "environment": "test-4"}

page1 = {"count": 4, "next": "search_2", "previous": None, "results": [item0, item1]}

page2 = {"count": 4, "next": "search_2", "previous": None, "results": [item2, item3]}


def processor(id):
    return {
        "id": id,
        "name": "test-4",
        "version": 1,
        "docker_image": "test-4",
        "environment": "test-4",
    }


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/test/" and not kwargs:
        return MockResponse(page1, url)

    if url == "https://api.refine.bio/v1/test/" and kwargs["params"] == {"offset": 2, "limit": 2}:
        return MockResponse(page2, url)

    if url == "https://api.refine.bio/v1/processors/" and kwargs["params"] == {"limit": 10}:
        return MockResponse(
            {
                "count": 20,
                "next": "foo",
                "previous": None,
                "results": [processor(i) for i in range(10)],
            },
            url,
        )

    if url == "https://api.refine.bio/v1/processors/" and kwargs["params"] == {
        "offset": 10,
        "limit": 10,
    }:
        return MockResponse(
            {
                "count": 20,
                "next": "foo",
                "previous": None,
                "results": [processor(i) for i in range(10, 20)],
            },
            url,
        )


class PaginatedListTests(unittest.TestCase, CustomAssertions):
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_indexing(self, mock_request):
        T = pyrefinebio.Processor
        response = MockResponse(page1, "https://api.refine.bio/v1/test/")
        paginatedList = pyrefinebio.util.create_paginated_list(T, response)

        self.assertObject(paginatedList[0], item0)
        self.assertObject(paginatedList[1], item1)
        self.assertObject(paginatedList[2], item2)
        self.assertObject(paginatedList[3], item3)

        self.assertObject(paginatedList[-1], item3)
        self.assertObject(paginatedList[-2], item2)

        with self.assertRaises(IndexError):
            paginatedList[4]

        with self.assertRaises(IndexError):
            paginatedList[-5]

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_iteration(self, mock_request):
        T = pyrefinebio.Processor
        response = MockResponse(page1, "https://api.refine.bio/v1/test/")
        paginatedList = pyrefinebio.util.create_paginated_list(T, response)

        actual = [item0, item1, item2, item3]

        i = 0

        for processor in paginatedList:
            self.assertObject(processor, actual[i])
            i += 1

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_slicing(self, mock_request):
        paginatedList = pyrefinebio.Processor.search(limit=10)

        actual = paginatedList[5:15]

        self.assertEqual(len(actual), 10)

        self.assertObject(actual[0], processor(5))
        self.assertObject(actual[1], processor(6))
        self.assertObject(actual[2], processor(7))
        self.assertObject(actual[3], processor(8))
        self.assertObject(actual[4], processor(9))
        self.assertObject(actual[5], processor(10))
        self.assertObject(actual[6], processor(11))
        self.assertObject(actual[7], processor(12))
        self.assertObject(actual[8], processor(13))
        self.assertObject(actual[9], processor(14))

        self.assertEqual(len(mock_request.call_args_list), 2)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_slicing_with_step(self, mock_request):
        paginatedList = pyrefinebio.Processor.search(limit=10)

        actual = paginatedList[5:15:2]

        self.assertEqual(len(actual), 5)

        self.assertObject(actual[0], processor(5))
        self.assertObject(actual[1], processor(7))
        self.assertObject(actual[2], processor(9))
        self.assertObject(actual[3], processor(11))
        self.assertObject(actual[4], processor(13))

        self.assertEqual(len(mock_request.call_args_list), 2)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_slicing_negative_indices(self, mock_request):
        paginatedList = pyrefinebio.Processor.search(limit=10)

        actual = paginatedList[5:-5]

        self.assertEqual(len(actual), 10)

        self.assertObject(actual[0], processor(5))
        self.assertObject(actual[1], processor(6))
        self.assertObject(actual[2], processor(7))
        self.assertObject(actual[3], processor(8))
        self.assertObject(actual[4], processor(9))
        self.assertObject(actual[5], processor(10))
        self.assertObject(actual[6], processor(11))
        self.assertObject(actual[7], processor(12))
        self.assertObject(actual[8], processor(13))
        self.assertObject(actual[9], processor(14))

        self.assertEqual(len(mock_request.call_args_list), 2)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_slicing_negative_step(self, mock_request):
        paginatedList = pyrefinebio.Processor.search(limit=10)

        actual = paginatedList[15:5:-1]

        self.assertEqual(len(actual), 10)

        self.assertObject(actual[0], processor(15))
        self.assertObject(actual[1], processor(14))
        self.assertObject(actual[2], processor(13))
        self.assertObject(actual[3], processor(12))
        self.assertObject(actual[4], processor(11))
        self.assertObject(actual[5], processor(10))
        self.assertObject(actual[6], processor(9))
        self.assertObject(actual[7], processor(8))
        self.assertObject(actual[8], processor(7))
        self.assertObject(actual[9], processor(6))

        self.assertEqual(len(mock_request.call_args_list), 2)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_slicing_default_start_stop(self, mock_request):
        paginatedList = pyrefinebio.Processor.search(limit=10)

        actual = paginatedList[::5]

        self.assertEqual(len(actual), 4)

        self.assertObject(actual[0], processor(0))
        self.assertObject(actual[1], processor(5))
        self.assertObject(actual[2], processor(10))
        self.assertObject(actual[3], processor(15))

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_slicing_default_start_stop_negative_step(self, mock_request):
        paginatedList = pyrefinebio.Processor.search(limit=10)

        actual = paginatedList[::-5]

        self.assertEqual(len(actual), 4)

        self.assertObject(actual[0], processor(19))
        self.assertObject(actual[1], processor(14))
        self.assertObject(actual[2], processor(9))
        self.assertObject(actual[3], processor(4))



import os
import unittest
from unittest.mock import patch

import pyrefinebio
from pyrefinebio.config import Config
from tests.custom_assertions import CustomAssertions
from tests.mocks import MockResponse

token = {"id": "123456789", "is_activated": False, "terms_and_conditions": "TL;DR"}

token2 = {"id": "test", "is_activated": True, "terms_and_conditions": "TL;DR"}


def mock_request(method, url, **kwargs):

    if url == "https://api.refine.bio/v1/token/":
        return MockResponse(token, url)

    if url == "https://api.refine.bio/v1/token/123456789/":
        return MockResponse(token, url)

    if url == "https://api.refine.bio/v1/token/test/":
        return MockResponse(token2, url)


class TokenTests(unittest.TestCase, CustomAssertions):
    def setUp(self):
        Config._instance = None

    @classmethod
    def tearDownClass(cls):
        if os.path.exists("./temp"):
            os.remove("./temp")

        os.environ.pop("REFINEBIO_CONFIG_FILE", None)

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_token_create(self, mock_request):
        result = pyrefinebio.Token(email_address="")
        self.assertEqual(result.id, token["id"])

    @patch("pyrefinebio.config.yaml.dump")
    @patch("pyrefinebio.config.open")
    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_token_save(self, mock_request, mock_open, mock_yaml):
        os.environ["REFINEBIO_CONFIG_FILE"] = "test"
        mock_open.return_value.__enter__.return_value = "file"

        token = pyrefinebio.Token(id="test")
        token.save_token()

        mock_open.assert_called_with("test", "w")
        mock_yaml.assert_called_with(
            {
                "token": token.id,
                "base_url": "https://api.refine.bio/v1/",
                "api_max_calls_per_second": 10,
            },
            "file",
        )

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_token_save_creates_file(self, mock_request):
        os.environ["REFINEBIO_CONFIG_FILE"] = "./temp"

        if os.path.exists("./temp"):
            os.remove("./temp")

        token = pyrefinebio.Token(id="test")
        token.save_token()

        self.assertTrue(os.path.exists("./temp"))

    def test_token_save_bad_id(self):
        with self.assertRaises(pyrefinebio.exceptions.BadRequest) as br:
            token = pyrefinebio.Token(id="test")
            token.save_token()

        self.assertEqual(
            br.exception.base_message,
            "Bad Request: "
            "Token with id 'test' does not exist in refine.bio. "
            "Please create a new token using pyrefinebio.Token().",
        )

    @patch("pyrefinebio.api_interface.requests.request", side_effect=mock_request)
    def test_token_save_unactivated(self, mock_request):
        with self.assertRaises(pyrefinebio.exceptions.BadRequest) as br:
            token = pyrefinebio.Token(id="123456789")
            token.save_token()

        self.assertEqual(
            br.exception.base_message,
            "Bad Request: "
            "Token with id '123456789' is not activated. "
            "Please activate your token with `agree_to_terms_and_conditions()` before saving it.",
        )

    @patch("pyrefinebio.token.put_by_endpoint")
    def test_token_load(self, mock_put):
        mock_put.return_value = True
        pyrefinebio.Token(id="this-is-a-test-token").agree_to_terms_and_conditions()

        token = pyrefinebio.Token.load_token()

        self.assertEqual(token.id, "this-is-a-test-token")
